"""Module provider for Mythic Beasts"""

import hashlib
import json
import logging
from argparse import ArgumentParser
from typing import List

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Mythic Beasts"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["mythic-beasts.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            There are two ways to provide an authentication granting access to the Mythic Beasts API
            1 - With your API credentials (user/password), using --auth-username and --auth-password flags.
            2 - With an API token, using --auth-token flags.
            These credentials and tokens must be generated using the Mythic Beasts API v2.
        """
        parser.add_argument(
            "--auth-username",
            help="specify API credentials username",
        )
        parser.add_argument(
            "--auth-password",
            help="specify API credentials password",
        )
        parser.add_argument(
            "--auth-token",
            help="specify API token for authentication",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.mythic-beasts.com/dns/v2"
        self.auth_token = None

    def authenticate(self):
        # may need to get auth token
        if self.auth_token is None and self._get_provider_option("auth_token") is None:
            auth_request = requests.request(
                "POST",
                "https://auth.mythic-beasts.com/login",
                data={"grant_type": "client_credentials"},
                auth=(
                    self._get_provider_option("auth_username"),
                    self._get_provider_option("auth_password"),
                ),
            )
            auth_request.raise_for_status()
            post_result = auth_request.json()

            if not post_result["access_token"]:
                raise AuthenticationError(
                    "Error, could not get access token "
                    f"for Mythic Beasts API for user: {self._get_provider_option('auth_username')}"
                )

            self.auth_token = post_result["access_token"]
        elif self.auth_token is None:
            self.auth_token = self._get_provider_option("auth_token")

        payload = self._get("/zones")

        if self.domain is None:
            if not payload["zones"]:
                raise AuthenticationError("No domain found")
            if len(payload["zones"]) > 1:
                raise AuthenticationError(
                    "Too many domains found. This should not happen"
                )
            else:
                self.domain = payload["zones"][0]
        else:
            if not payload["zones"]:
                raise AuthenticationError("No domain found")
            if self.domain not in payload["zones"]:
                raise AuthenticationError("Requested domain not found")

        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        LOGGER.debug("type %s", rtype)
        LOGGER.debug("name %s", name)
        LOGGER.debug("content %s", content)

        if rtype == "CNAME":
            content = self._fqdn_name(content)

        data = {
            "records": [
                {
                    "host": self._relative_name(name),
                    "type": rtype,
                    "data": content,
                }
            ]
        }
        if self._get_lexicon_option("ttl"):
            data["records"][0]["ttl"] = self._get_lexicon_option("ttl")

        payload = {"success": True}
        try:
            payload = self._post(f"/zones/{self.domain}/records", data)
        except requests.exceptions.HTTPError as err:
            if (
                err.response.status_code == 400
                and err.response.json()["errors"][0][0:16] == "Duplicate record"
            ):
                LOGGER.debug("create_record (ignored, duplicate)")
            else:
                raise

        if rtype == "A" or rtype == "TXT":
            # need to wait and poll here until verified that DNS change is live
            try:
                self._get(
                    f"/zones/{self.domain}/records/{self._relative_name(name)}/{rtype}?verify"
                )
            except requests.exceptions.HTTPError:
                LOGGER.debug("Timed out trying to verify changes were live")
                raise

        if "message" in payload:
            return payload["message"]
        elif "success" in payload:
            return payload["success"]

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        filter_obj = {}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["host"] = self._relative_name(name)
        if content:
            filter_obj["data"] = content

        records = []
        payload = self._get(f"/zones/{self.domain}/records", filter_obj)

        LOGGER.debug("payload: %s", payload)

        for record in payload["records"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["host"]),
                "ttl": record["ttl"],
                "content": record["data"],
                # no id is available, so we need to make our own
                "id": _identifier(record),
            }
            if record["type"] == "MX" and record["mx_priority"]:
                processed_record["options"] = {
                    "mx": {"priority": record["mx_priority"]}
                }

            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        LOGGER.debug("Number of records retrieved: %d", len(records))
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self.list_records(rtype, self._full_name(name))
            if len(records) == 1:
                matching_record = records[0]
                filter_obj = {}
                filter_obj["type"] = matching_record["type"]
                filter_obj["host"] = self._relative_name(matching_record["name"])

            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )
        else:
            records = self.list_records()
            for record in records:
                if record["id"] == identifier:
                    matching_record = record
                    break
            else:
                raise Exception("Can't find record with that id!")

            filter_obj = {}
            filter_obj["type"] = matching_record["type"]
            filter_obj["host"] = self._relative_name(matching_record["name"])
            filter_obj["data"] = matching_record["content"]

        data = {"records": [{}]}
        if rtype:
            data["type"] = rtype
        if name:
            data["host"] = self._relative_name(name)
        if content:
            data["records"][0]["data"] = content
        if self._get_lexicon_option("ttl"):
            data["records"][0]["ttl"] = self._get_lexicon_option("ttl")

        LOGGER.debug(data)

        payload = self._put(
            f"/zones/{self.domain}/records/{self._relative_name(matching_record['name'])}/{matching_record['type']}",
            data,
            filter_obj,
        )

        LOGGER.debug("update_record: %s", payload["message"])
        return payload["message"]

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        filter_obj = {}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["host"] = self._relative_name(name)
        if content:
            filter_obj["data"] = content

        records = self.list_records(rtype, name, content)

        for record in records:
            LOGGER.debug("delete_records: %s", record)
            name = self._relative_name(record["name"])
            rtype = record["type"]
            if identifier is not None and identifier != record["id"]:
                continue

            self._delete(f"/zones/{self.domain}/records/{name}/{rtype}", filter_obj)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        # When editing DNS zone, results are not live immediately.
        # In this case, call to verify will return 409 HTTP error.
        # We use the Retry extension to retry the requests until
        # we get a processable response (402 HTTP status, or an HTTP error != 409)
        retries = Retry(
            total=10,
            backoff_factor=0.5,
            status_forcelist=[409],
            allowed_methods=frozenset(["GET", "PUT", "POST", "DELETE", "PATCH"]),
        )

        session = requests.Session()
        session.mount("https://", HTTPAdapter(max_retries=retries))

        headers = {"Content-Type": "application/json"}
        headers["Authorization"] = f"Bearer {self.auth_token}"
        response = session.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()


# Return hash id for record
def _identifier(record):
    m = hashlib.sha256()
    m.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
    m.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
    m.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

    return m.hexdigest()[0:7]



import sys
import warnings

from lexicon._private.discovery import find_providers, load_provider_module

warnings.warn(
    """\
Package lexicon.providers is deprecated and will be removed in Lexicon 4>=.

Providers implementations are private, and are not meant to be used directly. Please use lexicon.client.Client instead.
Base provider interface is migrated to lexicon.interfaces.provider. Please use it instead to implement a new provider.
""",
    DeprecationWarning,
    stacklevel=2,
)

for module_name, available in find_providers().items():
    sys.modules[f"lexicon.providers.{module_name}"] = load_provider_module(module_name)



"""Module provider for Digital Ocean"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Digital Ocean"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["digitalocean.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify token for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.digitalocean.com/v2"

    def authenticate(self):
        self._get(f"/domains/{self.domain}")
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        # check if record already exists
        ttl = self._get_lexicon_option("ttl")

        if not self.list_records(rtype, name, content):
            record = {
                "type": rtype,
                "name": self._relative_name(name),
                "data": content,
                "ttl": ttl,
            }
            if rtype == "CNAME":
                # make sure a the data is always a FQDN for CNAMe.
                record["data"] = record["data"].rstrip(".") + "."

            self._post(f"/domains/{self.domain_id}/records", record)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        url = f"/domains/{self.domain_id}/records"
        records = []
        payload = {}

        next_url = url
        while next_url is not None:
            payload = self._get(next_url)
            if (
                "links" in payload
                and "pages" in payload["links"]
                and "next" in payload["links"]["pages"]
            ):
                next_url = payload["links"]["pages"]["next"]
            else:
                next_url = None

            for record in payload["domain_records"]:
                processed_record = {
                    "type": record["type"],
                    "name": f"{record['name']}.{self.domain_id}",
                    "ttl": record["ttl"],
                    "content": record["data"],
                    "id": record["id"],
                }
                records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["data"] = content

        ttl = self._get_lexicon_option("ttl")
        if ttl:
            data["ttl"] = ttl

        self._put(f"/domains/{self.domain_id}/records/{identifier}", data)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/domains/{self.domain_id}/records/{record_id}")

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._get_provider_option('auth_token')}",
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        response = requests.request(
            action,
            url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        return response.json()



"""Module provider for hostingde (Hosting.de)"""

import json
import logging
import time
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# be aware to provide an auth_token
# LEXICON_HOSTINGDE_AUTH_TOKEN


class Provider(BaseProvider):
    """Provider class for Hosting"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["hosting.de"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api key for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://secure.hosting.de/api/dns/v1/json"

    def authenticate(self):
        response = self._get_zone_config()

        LOGGER.debug("authenticate debug: %s", response)
        if response == []:
            raise Exception(f"Domain {self.domain} not found")

        self.domain_id = response.get("id", None)

    def cleanup(self) -> None:
        pass

    # Helper
    def _get_zone_config(self):
        data = {"filter": {"field": "ZoneName", "value": self.domain}}
        response = self._request(action="POST", url="/zoneConfigsFind", data=data)
        if response != []:
            return response[0]

        return []

    # Normal Behaviour List all records. If filters are provided, send to the API if possible,
    # else apply filter locally. Return value should be a list of records.
    # Return an empty list if no records found
    # type, name and content are used to filter records.
    # caused the fact, that provider will enclose TXT values with ""
    # we have filter content afterwords
    def list_records(self, rtype=None, name=None, content=None):
        data = {}
        subfilter = []  # used by API to filter
        subfilter.append({"field": "zoneConfigId", "value": self.domain_id})
        if rtype:
            subfilter.append({"field": "RecordType", "value": rtype})
        if name:
            subfilter.append({"field": "RecordName", "value": self._full_name(name)})
        if subfilter:
            data.update(
                {"filter": {"subFilterConnective": "AND", "subFilter": subfilter}}
            )

        LOGGER.debug("list_records filter: %s", data)
        raw_records = self._request(action="POST", url="/recordsFind", data=data)

        processed_records = []

        for record in raw_records:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "id": record["id"],
                "content": record["content"],
            }
            if record["priority"]:
                processed_record["priority"] = record["priority"]
            processed_record = self._clean_TXT_record(processed_record)

            processed_records.append(processed_record)

        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    # Normal Behavior Create a new DNS record. Return a boolean True if successful.
    # If Record Already Exists Do nothing. DO NOT throw exception.
    # TTL If not specified or set to 0, use reasonable default.
    def create_record(self, rtype, name, content):
        records = self.list_records(rtype, name, content)
        if records:
            LOGGER.debug("not creating duplicate record: %s", records[0])
            return True

        zone_config = self._get_zone_config()
        priority = self._get_lexicon_option("priority")
        ttl = self._get_lexicon_option("ttl")

        record = {"name": self._full_name(name), "type": rtype, "content": content}
        if ttl:
            if int(ttl) < 60:
                LOGGER.warning("ttl must be minimum 60")
                ttl = 60
            record["ttl"] = int(ttl)
        if priority:
            record["priority"] = int(priority)

        LOGGER.debug("create_record: %s", record)
        data = {"zoneConfig": zone_config, "recordsToAdd": [record]}

        self._request(action="POST", url="/zoneUpdate", data=data)
        return True

    # Normal Behaviour Update a record. Record to be updated can be specified by providing id OR
    # name, type and content. Return a boolean
    # True if successful.
    # TTL:
    #    If not specified, do not modify ttl.
    #    If set to 0, reset to reasonable default.
    # No Match Throw exception?
    # Update a record. use delete & create cause there is no API update call
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier:
            records = self.list_records()
            records = [r for r in records if r["id"] == identifier]
        else:
            records = self.list_records(rtype, name, None)

        if not records:
            raise Exception("Record not found")
        if len(records) > 1:
            raise Exception("Record not unique")
        orig_record = records[0]
        orig_id = orig_record["id"]

        new_rtype = rtype if rtype else orig_record["type"]
        new_name = name if name else orig_record["name"]
        new_content = content if content else orig_record["content"]

        self.delete_record(orig_id)
        return self.create_record(new_rtype, new_name, new_content)

    # Normal Behaviour Remove a record. Record to be deleted can be
    # specified by providing id OR name, type and content.
    # Return a boolean True if successful.
    # No Match Do nothing. DO NOT throw exception
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_ids)
        records = []
        for record_id in delete_record_ids:
            records.append({"id": record_id})

        zone_config = self._get_zone_config()
        data = {"zoneConfig": zone_config, "recordsToDelete": records}

        self._request(action="POST", url="/zoneUpdate", data=data)

        # sometimes it takes some time to delete a record
        # so, here check after deleting and loop
        # if record(s) is not deleted after 30 seconds > False
        retries = 30
        wait_time = int(self._get_provider_option("retry_wait_time") or 1000)
        for record_id in delete_record_ids:
            while True:
                if self.list_records(record_id):
                    if retries < 1:
                        break
                    retries = retries - 1
                    time.sleep(wait_time / 1000)
                else:
                    break

        LOGGER.debug("delete_records: %s", retries > 0)
        return retries > 0

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        data.update({"authToken": self._get_provider_option("auth_token")})

        read_page = 1
        return_data = []
        retries = 30
        wait_time = int(self._get_provider_option("retry_wait_time") or 1000)

        # in some situations, API uses pagination
        # here we check and if there is pagination, go through all pages
        while True:
            page_data = data
            page_data.update({"page": read_page})

            response = requests.request(
                action,
                self.api_endpoint + url,
                params=query_params,
                data=json.dumps(page_data),
                headers={"Content-Type": "application/json"},
            )
            response.raise_for_status()
            response_json = response.json()  # Work with json Object
            LOGGER.debug("_request response: %s", response_json)
            status = response_json.get("status", "*** no status available ***")
            # if API still busy just wait and retry
            if (
                status == "error"
                and response_json.get("errors")[0].get("value", "") == "blocked"
            ):
                if retries < 1:
                    raise Exception(f"Api error: {response_json.get('errors')}")
                retries = retries - 1
                time.sleep(wait_time / 1000)
                continue

            if status not in ("success", "pending"):
                raise Exception(f"Api error: {response_json.get('errors')}")
            # check if there is a data object
            read_data = response_json.get("response", {}).get("data", None)

            # if no data object, check if there is a records object
            if read_data is None:
                read_data = response_json.get("response", {}).get("records", None)

            # add response data to return data
            if read_data:
                return_data += read_data

            # is there pagination and more data available?
            total_pages = response_json.get("response", {}).get("totalPages", 0)
            if total_pages in (read_page, 0):
                break  # no more data

            read_page = read_page + 1

        return return_data



"""Module provider for Vercel"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    Implements the DNS Vercel provider.
    The API is quite simple: you can list all records, add one record or delete one record.
        - list is pretty straightforward: we get all records then filter for given parameters,
        - add uses directly the API to add a new record without any added complexity,
        - delete uses list + delete: we get the list of all records,
          filter on the given parameters and delete record by id,
        - update uses list + delete + add: we get the list of all records,
          find record for given identifier, then insert a new record and delete the old record.
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["vercel-dns.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            Vercel provider requires a token to access its API.
            You can generate one for your account on the following URL:
            https://vercel.com/account/tokens"""
        parser.add_argument("--auth-token", help="specify your API token")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.vercel.com"

    def authenticate(self):
        result = self._get(f"/v5/domains/{self.domain}")

        identifier = result.get("domain", {}).get("id")

        if not identifier:
            raise AuthenticationError(f"Error, domain {self.domain} not found")

        self.domain_id = identifier

    def cleanup(self) -> None:
        pass

    def list_records(self, rtype=None, name=None, content=None):
        result = self._get(f"/v4/domains/{self.domain}/records")

        raw_records = result["records"]
        if rtype:
            raw_records = [
                raw_record for raw_record in raw_records if raw_record["type"] == rtype
            ]
        if name:
            raw_records = [
                raw_record
                for raw_record in raw_records
                if raw_record["name"] == self._relative_name(name)
            ]
        if content:
            raw_records = [
                raw_record
                for raw_record in raw_records
                if raw_record["value"] == content
            ]

        records = []
        for raw_record in raw_records:
            records.append(
                {
                    "id": raw_record["id"],
                    "type": raw_record["type"],
                    "name": self._full_name(raw_record["name"]),
                    "content": raw_record["value"],
                    "ttl": raw_record["ttl"],
                }
            )

        LOGGER.debug("list_records: %s", records)

        return records

    def create_record(self, rtype, name, content):
        # We ignore creation if a record already exists for given rtype/name/content
        records = self.list_records(rtype, name, content)
        if records:
            LOGGER.debug("create_record (ignored, duplicate): %s", records[0]["id"])
            return True

        data = {"type": rtype, "name": self._relative_name(name), "value": content}

        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        result = self._post(f"/v2/domains/{self.domain}/records", data)

        if not result["uid"]:
            raise Exception("Error occured when inserting the new record.")

        LOGGER.debug("create_record: %s", result["uid"])

        return True

    def update_record(self, identifier, rtype=None, name=None, content=None):
        # Vercel do not allow to update a record, only add or remove.
        # So we get the corresponding record, dump or update
        # its content and insert it as a new record.
        # Then we remove the old record.
        records = []
        if identifier:
            records = self.list_records()
            records = [record for record in records if record["id"] == identifier]
        else:
            records = self.list_records(rtype, name)

        if not records:
            raise Exception(f"No record found for identifer: {identifier}")

        if len(records) > 1:
            LOGGER.warning(
                "Multiple records have been found for given parameters. "
                "Only first one will be updated (id: %s)",
                records[0]["id"],
            )

        data = {"type": rtype, "name": self._relative_name(name), "value": content}

        if not rtype:
            data["type"] = records[0]["type"]
        if not name:
            data["name"] = self._relative_name(records[0]["name"])
        if not content:
            data["value"] = records[0]["content"]
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        result = self._post(f"/v2/domains/{self.domain}/records", data)
        self._delete(f"/v2/domains/{self.domain}/records/{records[0]['id']}")

        LOGGER.debug("update_record: %s => %s", records[0]["id"], result["uid"])

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_ids)

        for delete_record_id in delete_record_ids:
            self._delete(f"/v2/domains/{self.domain}/records/{delete_record_id}")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        request = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Authorization": f"Bearer {self._get_provider_option('auth_token')}"
            },
        )

        request.raise_for_status()
        return request.json()



"""Main module of Lexicon. Defines the Client class, that holds all Lexicon logic."""

from __future__ import annotations

import logging
import os
import warnings
from contextlib import AbstractContextManager
from threading import local
from types import TracebackType
from typing import Any, Type

import dns.resolver
import tldextract

from lexicon import config as helper_config
from lexicon._private.discovery import find_providers as _find_providers
from lexicon._private.discovery import load_provider_module as _load_provider_module
from lexicon.exceptions import ProviderNotAvailableError
from lexicon.interfaces import Provider


class _ClientOperations:
    """
    Represents the entrypoint to execute several operations against the Client
    for a given resolved Provider already authenticated.
    """

    def __init__(self, provider: Provider):
        self.provider = provider

    def create_record(self, rtype: str, name: str, content: str) -> bool:
        """
        Create record. If record already exists with the same content, do nothing.
        """
        return self.provider.create_record(rtype, name, content)

    def list_records(
        self,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> list[dict[str, Any]]:
        """
        List all records. Return an empty list if no records found
        type, name and content are used to filter records.
        If possible filter during the query, otherwise filter after response is received.
        """
        return self.provider.list_records(rtype, name, content)

    def update_record(
        self,
        identifier: str | None = None,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> bool:
        """
        Update a record. Identifier must be specified.
        """
        return self.provider.update_record(identifier, rtype, name, content)

    def delete_record(
        self,
        identifier: str | None = None,
        rtype: str | None = None,
        name: str | None = None,
        content: str | None = None,
    ) -> bool:
        """
        Delete an existing record.
        If record does not exist, do nothing.
        If an identifier is specified, use it, otherwise do a lookup using type, name and content.
        """
        return self.provider.delete_record(identifier, rtype, name, content)


class Client(AbstractContextManager):
    """This is the Lexicon client, that will execute all the logic."""

    def __init__(
        self, config: helper_config.ConfigResolver | dict[str, Any] | None = None
    ):
        if not config:
            # If there is no config specified, we load a non-interactive configuration.
            self.config = (
                helper_config.ConfigResolver().with_env().with_config_dir(os.getcwd())
            )
        elif not isinstance(config, helper_config.ConfigResolver):
            # If config is not a ConfigResolver, we are in a legacy situation.
            # We protect this part of the Client API.
            self.config = helper_config.legacy_config_resolver(config)
        else:
            self.config = config

        domain = self.config.resolve("lexicon:domain")
        if not domain:
            raise AttributeError("domain")

        self._validate_provider()

        runtime_config = {}

        # Find the actual zone name for the domain
        if self.config.resolve("lexicon:resolve_zone_name") is not None:
            logging.debug(
                "Parameter resolve_zone_name is set, use dnspython to resolve the actual zone name"
            )
            zone_name = dns.resolver.zone_for_name(domain)
            runtime_config["domain"] = zone_name.to_text(omit_final_dot=True)
        else:
            logging.debug(
                "Parameter resolve_zone_name is not set, use tldextract to guess the zone name from known TLDs"
            )
            try:
                domain_extractor = tldextract.TLDExtract(
                    cache_dir=_resolve_tldextract_cache_path(),
                    include_psl_private_domains=True,
                )
            except TypeError:
                domain_extractor = tldextract.TLDExtract(
                    cache_file=_resolve_tldextract_cache_path(), include_psl_private_domains=True  # type: ignore
                )
            domain_parts = domain_extractor(domain)
            runtime_config["domain"] = f"{domain_parts.domain}.{domain_parts.suffix}"
        logging.debug(
            f"Actual zone name resolved for domain {domain}: {runtime_config['domain']}"
        )

        delegated = self.config.resolve("lexicon:delegated")
        if delegated:
            # handle delegated domain
            delegated = str(delegated).rstrip(".")
            initial_domain = str(runtime_config["domain"])
            if delegated != initial_domain:
                # convert to relative name
                if delegated.endswith(initial_domain):
                    delegated = delegated[: -len(initial_domain)]
                    delegated = delegated.rstrip(".")
                # update domain
                runtime_config["domain"] = f"{delegated}.{initial_domain}"
            logging.debug(
                f"Override resolved zone name because --delegated option is set: {runtime_config['domain']}"
            )

        self.provider_name = self.config.resolve(
            "lexicon:provider_name"
        ) or self.config.resolve("lexicon:provider")

        if not self.provider_name:
            raise ValueError("Could not resolve provider name.")

        self.config.add_config_source(helper_config.DictConfigSource(runtime_config), 0)

        provider_module = _load_provider_module(self.provider_name)
        self.provider_class: Type[Provider] = getattr(provider_module, "Provider")

        self._state = local()
        self._state.stack = []

    def __enter__(self) -> _ClientOperations:
        try:
            provider = self.provider_class(self.config)
            provider.authenticate()

            self._state.stack.append(provider)

            return _ClientOperations(provider)
        except Exception as e:
            self._state.stack.append(None)
            raise e

    def __exit__(
        self,
        __exc_type: type[BaseException] | None,
        __exc_value: BaseException | None,
        __traceback: TracebackType | None,
    ) -> bool | None:
        provider: Provider | None = self._state.stack.pop(-1)
        if provider:
            provider.cleanup()

        return None

    def execute(self) -> bool | list[dict[str, Any]]:
        """(deprecated) Execute provided configuration in class constructor to the DNS records"""
        message = """\
Method execute() is deprecated and will be removed in Lexicon 4>=.

Please remove action/type/name/content fields from Lexicon config,
and use the methods dedicated for each action (*_record()/list_records()).
These methods are available within the Lexicon client context manager.

Example for creating a record:

with Client(config) as operations:
    operations.create_record("TXT", "foo", "bar")
        """

        warnings.warn(message, DeprecationWarning, stacklevel=2)

        action = self.config.resolve("lexicon:action")
        identifier = self.config.resolve("lexicon:identifier")
        rtype = self.config.resolve("lexicon:type")
        name = self.config.resolve("lexicon:name")
        content = self.config.resolve("lexicon:content")

        if not action:
            raise AttributeError("action")
        if not rtype:
            raise AttributeError("type")

        try:
            executor = self.__enter__()

            if action == "create":
                if not name or not content:
                    raise ValueError("Missing record_type, name or content parameters.")
                return executor.create_record(rtype, name, content)

            if action == "list":
                return executor.list_records(rtype, name, content)

            if action == "update":
                return executor.update_record(identifier, rtype, name, content)

            if action == "delete":
                return executor.delete_record(identifier, rtype, name, content)

            raise ValueError(f"Invalid action statement: {action}")
        finally:
            self.__exit__(None, None, None)

    def _validate_provider(self) -> None:
        provider_name = self.config.resolve("lexicon:provider_name")
        if not provider_name:
            raise AttributeError("provider_name")

        try:
            available = _find_providers()[provider_name]
        except KeyError:
            raise ProviderNotAvailableError(
                f"This provider ({provider_name}) is not supported by Lexicon."
            )
        else:
            if not available:
                raise ProviderNotAvailableError(
                    f"This provider ({provider_name}) has required extra dependencies that are missing. "
                    f"Please run `pip install dns-lexicon[{provider_name}]` first before using it."
                )


def _resolve_tldextract_cache_path() -> str:
    if os.environ.get("TLDEXTRACT_CACHE_FILE"):
        logging.warning(
            "TLD_EXTRACT_CACHE_FILE environment variable is deprecated, please use TLDEXTRACT_CACHE_PATH instead."
        )
        os.environ["TLDEXTRACT_CACHE_PATH"] = os.environ["TLDEXTRACT_CACHE_FILE"]

    return os.path.expanduser(
        os.environ.get("TLDEXTRACT_CACHE_PATH", os.path.join("~", ".lexicon_tld_set"))
    )



"""Module provider for Hetzner"""

import json
import logging
from argparse import ArgumentParser
from time import sleep
from typing import Any, List, Optional, TypedDict, Union

import requests

from lexicon.config import ConfigResolver
from lexicon.exceptions import AuthenticationError, LexiconError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """
    Provider for Hetzner Cloud DNS at https://console.hetzner.com or
    https://api.hetzner.cloud; and Hetzner DNS at https://dns.hetzner.com for backwards-compatibility.

    Does not work for konsoleH or Domain Robot.

    If you're still using dns.hetzner.com, migration to Hetzner Cloud is recommended. See the [official migration guide](https://docs.hetzner.com/networking/dns/migration-to-hetzner-console/process).
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._hetzner_impl = self._decide_provider()

    @staticmethod
    def get_nameservers() -> list[str]:
        return ["ns.hetzner.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token", help="Specify Hetzner DNS or Cloud API token"
        )

    def authenticate(self) -> None:
        self._hetzner_impl.authenticate()
        self.domain_id = self._hetzner_impl.domain_id

    def create_record(self, rtype, name, content):
        return self._hetzner_impl.create_record(rtype, name, content)

    def list_records(
        self,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> list[dict[str, Any]]:
        return self._hetzner_impl.list_records(rtype, name, content)

    def update_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        return self._hetzner_impl.update_record(identifier, rtype, name, content)

    def delete_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        return self._hetzner_impl.delete_record(identifier, rtype, name, content)

    def _decide_provider(self) -> BaseProvider:
        """
        To provide backwards-compatibilty, implementation for Hetzner DNS (old API) is kept. When the old API is completely shut down, this can be removed and HetznerCloud can be made the only provider in this file.
        """
        token = self._get_provider_option("auth_token")
        if token is None:
            raise LexiconError("auth-token must be passed.")
        if len(token) == 32:
            return HetznerDns(self.config)
        return HetznerCloud(self.config)


class HetznerDns(BaseProvider):

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["ns.hetzner.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="Specify Hetzner DNS API token")

    def __init__(self, config):
        super(HetznerDns, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://dns.hetzner.com/api/v1"

    def authenticate(self):
        provider = self._get_zone_by_domain(self.domain)
        self.domain_id = provider["id"]

    def create_record(self, rtype, name, content):
        data = {
            # hetzner needs the FQDN name if it does not belong to the managed domain itself
            "name": self._get_record_name(self.domain, name),
            "type": rtype,
            "value": content,
            "zone_id": self.domain_id,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        records = self.list_records(rtype=rtype, name=name, content=content)
        if len(records) >= 1:
            for record in records:
                LOGGER.warning(
                    "Duplicate record %s %s %s with id %s",
                    rtype,
                    name,
                    content,
                    record["id"],
                )
            return True
        self._post("/records", data)
        return True

    def list_records(self, rtype=None, name=None, content=None):
        filter_obj = {"zone_id": self.domain_id}
        payload = self._get("/records", filter_obj)
        records = map(self._hetzner_record_to_lexicon_record, payload["records"])
        filtered_records = self._filter_records(
            records, rtype, name if name is not None else None, content
        )

        return filtered_records

    def _filter_records(self, records, rtype=None, name=None, content=None):
        return [
            record
            for record in records
            if (rtype is None or record["type"] == rtype)
            and (name is None or record["name"] == self._full_name(name))
            and (content is None or record["content"] == content)
        ]

    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {
            "type": rtype,
            "name": self._get_record_name(self.domain, name),
            "value": content,
            "zone_id": self.domain_id,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        update_identifier = identifier
        if update_identifier is None:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                update_identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type, name and content - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type, name and content - won't update"
                )
        self._put(f"/records/{update_identifier}", data)
        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if identifier is None:
            records = self.list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        for record_id in delete_record_ids:
            self._delete(f"/records/{record_id}")
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Auth-API-Token": self._get_provider_option("auth_token"),
                "Content-Type": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _get_zone_by_domain(self, domain):
        """
        Requests all dns zones from your Hetzner account and searches for a specific
        one to determine the ID of it
        :param domain: Name of domain for which dns zone should be searched
        :rtype: dict
        :return: The dictionary of the zone with ``domain`` in the 'name' key
        :raises Exception: If no zone was found
        :raises KeyError, ValueError: If the response is malformed
        :raises urllib.error.HttpError: If request to /zones did not return 200
        """
        filter_obj = {"name": domain}
        payload = self._get("/zones", filter_obj)
        zones = payload["zones"]
        for zone in zones:
            if zone["name"] == domain:
                return zone
        raise AuthenticationError(f"No zone was found in account matching {domain}")

    def _get_record_name(self, domain, record_name):
        """
        Get the name attribute appropriate for hetzner api. This means it's the name
        without domain name if record name ends with managed domain name else a fqdn
        :param domain: Name of domain for which dns zone should be searched
        :param record_name: The record name to convert
        :rtype: str
        :return: The record name in an appropriate format for hetzner api
        """
        if record_name.rstrip(".").endswith(domain):
            record_name = self._relative_name(record_name)
        return record_name

    @staticmethod
    def _pretty_json(data):
        return json.dumps(data, sort_keys=True, indent=4, separators=(",", ": "))

    def _hetzner_record_to_lexicon_record(self, hetzner_record):
        lexicon_record = {
            "id": hetzner_record["id"],
            "name": self._full_name(hetzner_record["name"]),
            "content": hetzner_record["value"],
            "type": hetzner_record["type"],
        }
        if "ttl" in hetzner_record:
            lexicon_record["ttl"] = hetzner_record["ttl"]
        return lexicon_record


Record = TypedDict("Record", {"value": str})
RecordSet = TypedDict(
    "RecordSet",
    {
        "name": str,
        "type": str,
        "ttl": int,
        "records": list[Record],
    },
)


class HetznerCloud(BaseProvider):
    API_ENDPOINT = "https://api.hetzner.cloud/v1/zones"

    @staticmethod
    def get_nameservers() -> list[str]:
        return [
            "hydrogen.ns.hetzner.com",
            "oxygen.ns.hetzner.com",
            "helium.ns.hetzner.de",
        ]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="Specify Hetzner DNS API token")

    def __init__(self, config: Union[ConfigResolver, dict[str, Any]]):
        super(HetznerCloud, self).__init__(config)
        self.domain_id = None

    def authenticate(self) -> None:
        self.domain_id = self._fetch_zone(self.domain)["id"]

    def create_record(self, rtype: str, name: str, content: str) -> bool:
        duplicate_records = self.list_records(rtype, name, content)
        if len(duplicate_records) > 0:
            LOGGER.info(f"Record {rtype} {name} {content} already exists")
            return True

        action = self._post(
            f"{self._rrset_url(name, rtype)}/actions/add_records",
            {"ttl": self._get_ttl(), "records": self._records_from(rtype, content)},
        )["action"]

        return self._wait_for_action(action)

    def list_records(
        self,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> list[dict[str, Any]]:
        record_sets = self._get_record_sets(rtype, name)

        name = self._full_name(name) if name else None
        return [
            record
            for record_set in record_sets
            for record in self._rrset_to_records(record_set)
            if (rtype is None or record["type"] == rtype)
            and (name is None or record["name"] == name)
            and (content is None or record["content"] == content)
        ]

    def update_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if rtype is None or name is None or content is None:
            raise LexiconError(
                "rtype, name and content need to be set in order to update a record."
            )

        if identifier:
            raise LexiconError("Hetzner API does not provide ids per record")

        records = self.list_records(rtype, name)
        if len(records) < 1:
            raise Exception("No records found matching type and name - won't update")
        elif len(records) > 1:
            raise Exception(
                "Multiple records found matching type and name - won't update"
            )

        action = self._post(
            f"{self._rrset_url(name, rtype)}/actions/set_records",
            {"records": self._records_from(rtype, content)},
        )["action"]
        if not self._wait_for_action(action):
            return False

        # verify the ttl is correctly set
        ttl = self._get_ttl()
        if ttl and records[0]["ttl"] is not ttl:
            return self._wait_for_action(
                self._post(
                    f"{self._rrset_url(name, rtype)}/actions/change_ttl", {"ttl": ttl}
                )["action"]
            )
        return True

    def delete_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if rtype is None or name is None:
            raise LexiconError("both rtype and name need to be passed.")

        if identifier:
            raise LexiconError("Hetzner API does not provide ids per record")

        if content is None:
            # Entire record set should be deleted
            action = self._delete(self._rrset_url(name, rtype))["action"]
            return self._wait_for_action(action)
        else:
            # Record should be taken out of set
            action = self._post(
                f"{self._rrset_url(name, rtype)}/actions/remove_records",
                {"records": self._records_from(rtype, content)},
            )["action"]

            return self._wait_for_action(action)

    # Helpers
    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: Optional[dict[str, Any]] = None,
        query_params: Optional[dict[str, Any]] = None,
    ):
        query_params = query_params or {}

        headers = {
            "Authorization": f"Bearer {self._get_provider_option('auth_token')}",
            "Accept": "application/json",
        }
        json_data = None
        if data:
            headers["Content-Type"] = "application/json"
            json_data = json.dumps(data)

        response = requests.request(
            action,
            self.API_ENDPOINT + url,
            params=query_params,
            data=json_data,
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _get_action(self, action_id):
        return self._get(f"/actions/{action_id}")["action"]

    def _is_action_running(self, action):
        return action["status"] == "running"

    def _was_action_successful(self, action):
        return action["status"] == "success"

    def _wait_for_action(self, action):
        if not self._is_action_running(action):
            return self._was_action_successful(action)

        wait_time = int(self._get_provider_option("retry_wait_time") or 500)
        while self._is_action_running(action):
            sleep(wait_time / 1000)
            action = self._get_action(action["id"])

        return self._was_action_successful(action)

    def _fetch_zone(self, domain: str) -> dict[str, Any]:
        try:
            return self._get(f"/{domain}")["zone"]
        except requests.HTTPError as err:
            if err.response.status_code == 401:
                raise AuthenticationError() from err
            elif err.response.status_code == 404:
                raise LexiconError(f"There is no zone for {domain}.") from err
            else:
                raise LexiconError(err) from err

    def _get_ttl(self) -> Optional[int]:
        ttl_str = self._get_lexicon_option("ttl")
        if not ttl_str:
            return None
        ttl = int(ttl_str)
        if not ttl:
            return None
        if ttl < 60 or ttl > 2147483647:
            raise LexiconError("TTL has to be between 60 and 2147483647")
        return ttl

    def _zone_url(self) -> str:
        return f"/{self.domain_id}"

    def _rrset_url(self, name: str, rtype: str) -> str:
        rrset_name = self._relative_name(name)
        return f"{self._zone_url()}/rrsets/{rrset_name}/{rtype}"

    def _rrset_to_records(self, rrset: RecordSet) -> list[dict[str, Any]]:
        return [
            {
                "id": None,
                "name": self._full_name(rrset["name"]),
                "content": self._get_content_from_record(
                    rrset["type"], record["value"]
                ),
                "type": rrset["type"],
                "ttl": rrset["ttl"],
            }
            for record in rrset["records"]
        ]

    def _get_content_from_record(self, rtype: str, content: str):
        if rtype == "TXT":
            return (
                content.removeprefix('"')
                .removesuffix('"')
                .replace('" "', "")
                .replace('\\"', '"')
            )
        return content

    def _records_from(self, rtype: str, content: str) -> list[Record]:
        if rtype == "TXT":
            escaped_content = content.replace('"', '\\"')

            parts = []
            for start in range(0, len(escaped_content), 255):
                end = min(start + 255, len(escaped_content))
                parts.append('"' + escaped_content[start:end] + '"')
            content = " ".join(parts)

        return [{"value": content}]

    def _get_record_sets(self, rtype, name):
        response = self._get(
            f"{self._zone_url()}/rrsets",
            {"type": rtype, "name": self._relative_name(name) if name else None},
        )
        record_sets: list[RecordSet] = response["rrsets"]

        # get paged rrsets
        while (
            response["meta"]["pagination"]["page"]
            < response["meta"]["pagination"]["last_page"]
        ):
            response = self._get(
                f"{self._zone_url()}/rrsets",
                {
                    "type": rtype,
                    "name": name,
                    "page": response["meta"]["pagination"]["next_page"],
                },
            )
            record_sets += response["rrsets"]

        return record_sets



"""Module provider for Timeweb"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

# Timeweb API does not support TTL. All records get 600 by default.
DEFAULT_TTL = 600
DEFAULT_PAGE_SIZE = 100


class Provider(BaseProvider):
    """
    Implements Timeweb DNS provider: https://timeweb.cloud.
    See also API documentation: https://timeweb.cloud/api-docs#tag/Domeny
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["timeweb.cloud"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token",
            help="specify API token for authentication",
        )

    @staticmethod
    def _check_ok(result) -> bool:
        error_code = result.get("error_code")
        if error_code:
            status_code = result["status_code"]
            message = result["message"]
            LOGGER.error(f"Status code: {status_code}. {message}")
            return False
        else:
            return True

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.timeweb.cloud/api/v1"
        self.api_dns = f"/domains/{self.domain}"
        self.api_dns_records = f"{self.api_dns}/dns-records"

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        else:
            LOGGER.debug(f"Preparing data: {data}")
        if query_params is None:
            query_params = {}
        else:
            LOGGER.debug(f"Preparing query params: {query_params}")
        full_url = self.api_endpoint + url
        LOGGER.debug(f"Sending {action} {full_url}")
        response = requests.request(
            action,
            full_url,
            params=query_params,
            data=json.dumps(data),
            headers={
                "Authorization": f'Bearer {self._get_provider_option("auth_token")}',
                "Content-Type": "application/json",
            },
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        r_json = response.json() if response.content else {}
        LOGGER.debug(f"Result: {r_json}")
        return r_json

    def _get_subdomain(self, record_name):
        if record_name.rstrip(".").endswith(self.domain):
            record_name = self._relative_name(record_name)
        return record_name

    def authenticate(self):
        result = self._get(self.api_dns)
        if self._check_ok(result):
            self.domain_id = result["domain"]["id"]
        else:
            raise AuthenticationError(f"Domain {self.domain} not found.")

    def create_record(self, rtype, name, content):
        data = {
            "subdomain": self._get_subdomain(name),
            "type": rtype,
            "value": content,
        }
        try:
            result = self._post(self.api_dns_records, data)
            return self._check_ok(result)
        except requests.exceptions.HTTPError as e:
            status_code = e.response.status_code
            if status_code == 409:
                LOGGER.warning(
                    f"Record {rtype} {name} returns {status_code} (already exists?)"
                )
                return True
            raise e

    def list_records(self, rtype=None, name=None, content=None):
        records = []
        processed_records = 0
        while True:
            query_params = {"limit": DEFAULT_PAGE_SIZE, "offset": processed_records}
            result = self._get(self.api_dns_records, query_params=query_params)
            if not self._check_ok(result):
                break
            for record in result["dns_records"]:
                r_rtype = record["type"]
                r_name = record["data"].get("subdomain", self.domain)
                r_content = record["data"]["value"]
                processed_record = {
                    "type": r_rtype,
                    "name": self._full_name(name) if name else r_name,
                    "ttl": DEFAULT_TTL,
                    "content": r_content,
                    "id": record["id"],
                }
                if (
                    (rtype is None or r_rtype == rtype)
                    and (name is None or r_name == self._get_subdomain(name))
                    and (content is None or r_content == content)
                ):
                    records.append(processed_record)
                processed_records += 1

            total_records = result["meta"]["total"]
            if processed_records >= total_records:
                break

        LOGGER.debug(f"list_records: {records}")
        LOGGER.debug(f"Number of records retrieved: {len(records)}")
        return records

    def update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                record = records[0]
                identifier = str(record["id"])
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )
        else:
            records = self.list_records()
            record = next((r for r in records if str(r["id"]) == identifier), None)

        data = {
            "subdomain": self._get_subdomain(name) if name else record["name"],
            "type": rtype if rtype else record["type"],
            "value": content if content else record["content"],
        }
        result = self._patch(f"{self.api_dns_records}/{identifier}", data)

        LOGGER.debug(f"update_record: {result}")
        return self._check_ok(result)

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if name:
            name = self._get_subdomain(name)
        if not identifier:
            records = self.list_records(rtype, name, content)
            ids_to_delete = [r["id"] for r in records]
        else:
            ids_to_delete = [identifier]

        if ids_to_delete:
            overall_ok = True
            for id_to_delete in ids_to_delete:
                result = self._delete(f"{self.api_dns_records}/{id_to_delete}")
                if self._check_ok(result):
                    LOGGER.debug(f"Successfully deleted record {id_to_delete}")
                else:
                    overall_ok = False
            return overall_ok
        else:
            LOGGER.warning("delete_record: no record found")
            return False



"""Module provider for DNS Simple"""

import json
import logging
import sys
from argparse import ArgumentParser
from typing import List

import requests

if sys.version_info[:2] < (3, 10):
    from importlib_metadata import Distribution, PackageNotFoundError
else:
    from importlib.metadata import Distribution, PackageNotFoundError

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


def _fetch_lexicon_version() -> str:
    """Retrieve current Lexicon version"""
    try:
        return Distribution.from_name("dns-lexicon").version
    except PackageNotFoundError:
        return "unknown"


class Provider(BaseProvider):
    """Provider class for DNS Simple"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dnsimple.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api token for authentication")
        parser.add_argument(
            "--auth-username", help="specify email address for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )
        parser.add_argument(
            "--auth-2fa",
            help="specify two-factor auth token (OTP) to use with email/password authentication",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.account_id = None
        self.api_endpoint = (
            self._get_provider_option("api_endpoint") or "https://api.dnsimple.com/v2"
        )
        # Lexicon version
        self._lexicon_version = _fetch_lexicon_version()

    def authenticate(self):
        payload = self._get("/accounts")

        if not payload[0]["id"]:
            raise AuthenticationError("No account id found")

        for account in payload:
            if account["plan_identifier"] is None:
                logging.warning(
                    "Skipping unconfigured account %s (%d). "
                    "To use this account, you must select a plan.",
                    account["email"],
                    account["id"],
                )
                continue

            dompayload = self._get(
                f"/{account['id']}/domains", query_params={"name_like": self.domain}
            )
            if dompayload and dompayload[0]["id"]:
                self.account_id = account["id"]
                self.domain_id = dompayload[0]["id"]
                break
        else:
            raise AuthenticationError(f"No domain found like {self.domain}")

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing

    def create_record(self, rtype, name, content):
        # check if record already exists
        existing_records = self.list_records(rtype, name, content)
        if len(existing_records) == 1:
            return True

        record = {"type": rtype, "name": self._relative_name(name), "content": content}
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            record["priority"] = self._get_lexicon_option("priority")
        if self._get_provider_option("regions"):
            record["regions"] = self._get_provider_option("regions")

        payload = self._post(f"/{self.account_id}/zones/{self.domain}/records", record)

        LOGGER.debug("create_record: %s", "id" in payload)
        return "id" in payload

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        filter_query = {}
        if rtype:
            filter_query["type"] = rtype
        if name:
            filter_query["name"] = self._relative_name(name)
        payload = self._get(
            f"/{self.account_id}/zones/{self.domain}/records", query_params=filter_query
        )

        records = []
        for record in payload:
            processed_record = {
                "type": record["type"],
                "name": (
                    f"{self.domain}"
                    if record["name"] == ""
                    else f"{record['name']}.{self.domain}"
                ),
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            if record["priority"]:
                processed_record["priority"] = record["priority"]
            records.append(processed_record)

        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {}

        if identifier is None:
            records = self.list_records(rtype, name, content)
            identifiers = [record["id"] for record in records]
        else:
            identifiers = [identifier]

        if name:
            data["name"] = self._relative_name(name)
        if content:
            data["content"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            data["priority"] = self._get_lexicon_option("priority")
        if self._get_provider_option("regions"):
            data["regions"] = self._get_provider_option("regions")

        for one_identifier in identifiers:
            self._patch(
                f"/{self.account_id}/zones/{self.domain}/records/{one_identifier}", data
            )
            LOGGER.debug("update_record: %s", one_identifier)

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/{self.account_id}/zones/{self.domain}/records/{record_id}")

        # is always True at this point; if a non 2xx response is returned, an error is raised.
        LOGGER.debug("delete_record: True")
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "User-Agent": f"lexicon/{self._lexicon_version} dnsimple",
        }
        default_auth = None

        if self._get_provider_option("auth_token"):
            default_headers["Authorization"] = (
                f"Bearer {self._get_provider_option('auth_token')}"
            )
        elif self._get_provider_option("auth_username") and self._get_provider_option(
            "auth_password"
        ):
            default_auth = (
                self._get_provider_option("auth_username"),
                self._get_provider_option("auth_password"),
            )
            if self._get_provider_option("auth_2fa"):
                default_headers["X-Dnsimple-OTP"] = self._get_provider_option(
                    "auth_2fa"
                )
        else:
            raise Exception("No valid authentication mechanism found")

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if response.text and response.json()["data"] is None:
            raise Exception("No data returned")

        return response.json()["data"] if response.text else None

    def _patch(self, url="/", data=None, query_params=None):
        return self._request("PATCH", url, data=data, query_params=query_params)



import sys
import warnings
from types import ModuleType

from lexicon._private import cli as _cli
from lexicon._private import discovery as _discovery
from lexicon._private import parser as _parser


class DeprecatedModule(ModuleType):
    def __init__(self, module: ModuleType, name: str):
        super().__init__(name)
        self._module = module

    def __getattr__(self, item):
        warnings.warn(
            f"Module {self.__name__} is deprecated and will be removed in Lexicon 4>=",
            DeprecationWarning,
            stacklevel=2,
        )
        return getattr(self._module, item)


sys.modules["lexicon.cli"] = DeprecatedModule(_cli, "lexicon.cli")
sys.modules["lexicon.parser"] = DeprecatedModule(_parser, "lexicon.parser")
sys.modules["lexicon.discovery"] = DeprecatedModule(_discovery, "lexicon.discovery")



"""Module provider for DNSPod"""

import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for DNSPod"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["dnsapi.cn"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-username", help="specify api id for authentication")
        parser.add_argument("--auth-token", help="specify token for authentication")

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://dnsapi.cn"

    def authenticate(self):
        payload = self._post("/Domain.Info", {"domain": self.domain})

        if payload["status"]["code"] != "1":
            raise AuthenticationError(payload["status"]["message"])

        self.domain_id = payload["domain"]["id"]

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, rtype, name, content):
        record = {
            "domain_id": self.domain_id,
            "sub_domain": self._relative_name(name),
            "record_type": rtype,
            "record_line": "\u9ed8\u8ba4",
            "value": content,
        }
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        payload = self._post("/Record.Create", record)

        if payload["status"]["code"] not in ["1", "31"]:
            raise Exception(payload["status"]["message"])

        LOGGER.debug("create_record: %s", payload["status"]["code"] == "1")
        return payload["status"]["code"] == "1"

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        payload = self._post("/Record.List", {"domain": self.domain})
        LOGGER.debug("payload: %s", payload)
        records = []
        for record in payload["records"]:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["value"],
                # this id is useless unless your doing record linking.
                # Lets return the original record identifier.
                "id": record["id"],
            }
            records.append(processed_record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        data = {
            "domain_id": self.domain_id,
            "record_id": identifier,
            "sub_domain": self._relative_name(name),
            "record_type": rtype,
            "record_line": "\u9ed8\u8ba4",
            "value": content,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")
        LOGGER.debug("data: %s", data)
        payload = self._post("/Record.Modify", data)
        LOGGER.debug("payload: %s", payload)
        if payload["status"]["code"] != "1":
            raise Exception(payload["status"]["message"])

        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._post(
                "/Record.Remove", {"domain_id": self.domain_id, "record_id": record_id}
            )

            # if payload['status']['code'] != '1':
            #    raise Exception(payload['status']['message'])

        # is always True at this point, if a non 200 response is returned an error is raised.
        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        data["login_token"] = (
            self._get_provider_option("auth_username")
            + ","
            + self._get_provider_option("auth_token")
        )
        data["format"] = "json"
        if query_params is None:
            query_params = {}
        default_headers = {}
        default_auth = None
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=default_headers,
            auth=default_auth,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()



"""Module provider for Hover"""

import json
import logging
import re
from argparse import ArgumentParser
from typing import List

import pyotp
import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Hover"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["hover.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )
        parser.add_argument(
            "--auth-totp-secret",
            help="specify base32-encoded shared secret to generate an OTP for authentication",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://www.hover.com/api"
        self.cookies = {}
        shared_secret = re.sub(
            r"\s*", "", self._get_provider_option("auth_totp_secret") or ""
        )
        self.totp = pyotp.TOTP(shared_secret)

    def authenticate(self) -> None:
        # Getting required cookies "hover_session" and "hoverauth"
        response = requests.get("https://www.hover.com/signin")
        self.cookies["hover_session"] = response.cookies["hover_session"]

        # Part one, login credentials
        payload = {
            "username": self._get_provider_option("auth_username"),
            "password": self._get_provider_option("auth_password"),
        }
        response = requests.post(
            "https://www.hover.com/signin/auth.json", json=payload, cookies=self.cookies
        )
        response.raise_for_status()

        # Part two, 2fa
        payload = {"code": self.totp.now()}
        response = requests.post(
            "https://www.hover.com/signin/auth2.json",
            json=payload,
            cookies=self.cookies,
        )
        response.raise_for_status()

        if "hoverauth" not in response.cookies:
            raise Exception("Unexpected auth response")
        self.cookies["hoverauth"] = response.cookies["hoverauth"]

        # Make sure domain exists
        # domain is stored in self.domain from BaseProvider

        domains = self._list_domains()
        for domain in domains:
            if domain["name"] == self.domain:
                self.domain_id = domain["id"]
                break
        else:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def cleanup(self) -> None:
        pass

    def _list_domains(self):
        response = self._get("/domains")

        domains = []
        for domain in response["domains"]:
            processed_domain = {
                "name": domain["domain_name"],
                "id": domain["id"],
                "active": (domain["status"] == "active"),
            }
            domains.append(processed_domain)

        LOGGER.debug("list_domains: %s", domains)
        return domains

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        payload = self._get(f"/control_panel/dns/{self.domain}")

        # payload['domains'] should be a list of len 1
        try:
            raw_records = payload["domain"]["dns"]
        except (KeyError, IndexError):
            raise Exception("Unexpected response")

        processed_records = []
        for record in raw_records:
            processed_record = {
                "type": record["type"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["content"],
                "id": record["id"],
            }
            processed_records.append(processed_record)

        if rtype:
            processed_records = [
                record for record in processed_records if record["type"] == rtype
            ]
        if name:
            name = self._relative_name(name)
            processed_records = [
                record for record in processed_records if name in record["name"]
            ]
        if content:
            processed_records = [
                record
                for record in processed_records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    def create_record(self, rtype, name, content):
        name = self._relative_name(name)
        records = self.list_records(rtype, name, content)
        if records:
            LOGGER.debug("not creating duplicate record: %s", records[0])
            return True

        record = {"name": name, "type": rtype, "content": content}
        if self._get_lexicon_option("ttl"):
            record["ttl"] = str(self._get_lexicon_option("ttl"))

        LOGGER.debug("create_record: %s", record)
        payload = {"id": f"domain-{self.domain}", "dns_record": record}
        response = self._post("/control_panel/dns", payload)

        return response["succeeded"]

    # Update a record. Hover cannot update name so we delete and recreate.
    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            records = self.list_records()
            records = [r for r in records if r["id"] == identifier]
        else:
            records = self.list_records(rtype, name, None)

        if not records:
            raise Exception("Record not found")
        if len(records) > 1:
            raise Exception("Record not unique")
        orig_record = records[0]
        orig_id = orig_record["id"]

        new_rtype = rtype if rtype else orig_record["type"]
        new_name = name if name else orig_record["name"]
        new_content = content if content else orig_record["content"]

        self.delete_record(orig_id)
        return self.create_record(new_rtype, new_name, new_content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_ids = [record["id"] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_ids)
        payload = {
            "domains": [
                {"id": f"domain-{self.domain}", "dns_records": delete_record_ids}
            ]
        }
        self._request("DELETE", "/control_panel/dns", payload)

        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        if not data:
            response = requests.request(
                action,
                self.api_endpoint + url,
                params=query_params,
                cookies=self.cookies,
                headers={"Content-Type": "application/json"},
            )
        else:
            response = requests.request(
                action,
                self.api_endpoint + url,
                params=query_params,
                data=json.dumps(data),
                cookies=self.cookies,
                headers={"Content-Type": "application/json"},
            )

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:  # response is not json
            raise Exception("Did not get JSON response.")



import requests

from lexicon.interfaces import Provider as BaseProvider

_ZONES_API = "https://api.hosting.ionos.com/dns/v1/zones"


class Provider(BaseProvider):

    @staticmethod
    def get_nameservers():
        return ["ui-dns.com", "ui-dns.org", "ui-dns.de", "ui-dns.biz"]

    @staticmethod
    def configure_parser(parser):
        parser.add_argument(
            "--api-key",
            help="IONOS api key: public prefix + period + key proper",
        )

    def authenticate(self):
        zones = self._get(_ZONES_API)
        for zone in zones:
            if zone["name"] == self.domain:
                self.domain_id = zone["id"]
                return
        raise Exception("domain not found: " + self.domain)

    def create_record(self, rtype, name, content):
        self._post(
            _ZONES_API + "/" + self.domain_id + "/records",
            data=[
                {
                    "name": self._full_name(name),
                    "type": rtype,
                    "content": content,
                    "ttl": self._get_lexicon_option("ttl"),
                    "prio": 0,
                    "disabled": False,
                }
            ],
        )
        return True

    def list_records(self, rtype=None, name=None, content=None):
        query_params = {}
        if rtype:
            query_params["recordType"] = rtype
        if name:
            query_params["recordName"] = self._full_name(name)
        data = self._get(_ZONES_API + "/" + self.domain_id, query_params)
        records = data["records"]
        records = [
            {
                "type": r["type"],
                "name": r["name"],
                "ttl": r["ttl"],
                "content": r["content"],
                "id": r["id"],
            }
            for r in records
        ]
        for record in records:
            self._clean_TXT_record(record)
        if content:
            records = [r for r in records if r["content"] == content]
        return records

    def update_record(self, identifier, rtype, name, content):
        self.delete_record(identifier, rtype, name, None)
        return self.create_record(rtype, name, content)

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            identifiers = [identifier]
        else:
            identifiers = [r["id"] for r in self.list_records(rtype, name, content)]
        for identifier in identifiers:
            self._delete(_ZONES_API + "/" + self.domain_id + "/records/" + identifier)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        response = requests.request(
            action,
            url,
            params=query_params,
            json=data,
            headers={
                "accept": "application/json",
                "x-api-key": self._get_provider_option("api_key"),
            },
        )
        response.raise_for_status()
        try:
            return response.json()
        except requests.exceptions.JSONDecodeError:
            return True



"""Lexicon exceptions module"""


class LexiconError(Exception):
    """
    Base Class for the Lexicon Exception hierarchy
    """


class ProviderNotAvailableError(LexiconError):
    """
    Custom exception to raise when a provider is not available,
    typically because some optional dependencies are missing
    """


class AuthenticationError(LexiconError):
    """
    Authentication to the provider failed, likely username, password or domain mismatch
    """



"""Module provider for Namecheap"""

import logging
import sys
from argparse import ArgumentParser
from typing import Dict, List, Optional, Tuple
from xml.etree.ElementTree import Element, fromstring

import requests
import tldextract

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

_NAMESPACE = "http://api.namecheap.com/xml.response"


class Provider(BaseProvider):
    """
    ========================================================================
    WARNING
    ========================================================================
    The Namecheap API does not add/update/delete a host but "gets" and
    "sets" ALL hosts at once (a complete replacement). In their comments on
    the API docs (https://www.namecheap.com/support/api/methods/domains-dns/set-hosts.aspx)
    it appears that not all host types are handled by their system.

    Their API only handles `[A, AAAA, CNAME, MX, MXE, TXT, URL, URL301, FRAME]`

    If you have SRV record, it may get lost. Also records configured as
    `A + DDNS` on their control panel will be downgrated to `A` records.
    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["namecheap.com", "registrar-servers.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument("--auth-token", help="specify api token for authentication")

        # earlier versions of the API expected the email address here
        # now they appear to want the username.
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )

        # FIXME What is the client IP used for?
        # this doesn't seem to be used for anything, but is required by their API
        # namecheap requires API requests to come from whitelisted domains, and this
        # is probably updated with the actual IP on their end.
        parser.add_argument(
            "--auth-client-ip",
            help="Client IP address to send to Namecheap API calls",
            default="127.0.0.1",
        )
        parser.add_argument(
            "--auth-sandbox",
            help="Whether to use the sandbox server",
            action="store_true",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.client = _Api(
            api_user=self._get_provider_option("auth_username") or "",
            api_key=self._get_provider_option("auth_token") or "",
            username=self._get_provider_option("auth_username") or "",
            client_ip=self._get_provider_option("auth_client_ip") or "",
            sandbox=self._get_provider_option("auth_sandbox") or False,
        )
        self.domain = self.domain
        self.domain_id = None

    def authenticate(self):
        """
        The Namecheap API is a little difficult to work with.
        Originally this method called PyNamecheap's `domains_getList`, which is
        connected to an API endpoint that only lists domains registered under
        the authenticating account. However, an authenticating Namecheap user
        may be permissioned to manage the DNS of additional domains. Namecheap's
        API does not offer a way to list these domains.

        This approach to detecting permissioned relies on some implementation
        details of the Namecheap API and the PyNamecheap module:

        * If the user does not own the domain, or is not permissioned to manage
          it in any way, Namecheap will return an error status, which
          PyNamecheap will instantly catch and raise.
        * If a non-error repsonse is returned, the XML payload is analyzed.
          If the user owns the domain it immediately returns valid. Otherwise
          we look for "All" Modification rights, or the hosts-edit permission.

        This is not feature complete and most-likely misses multiple scenarios
        where:
        * a user is privileged to manage the domain DNS, but via another "right"
        * a user is privileged to manage the domain, but DNS is not configured

        Important Note:
        * the Namecheap API has inconsistent use of capitalization with strings
          and a case-insensitive match should be made. e.g. the following appear
          in a payload: `False` and 'false', 'OK' and 'Ok'.

        TODO:
        * check payload for PremiumDNS
          <PremiumDnsSubscription>
            <IsActive>false</IsActive>
          </PremiumDnsSubscription>
        * check payload for other types of DNS
          <DnsDetails ProviderType="FREE" IsUsingOurDNS="true" HostCount="5"
                EmailType="No Email Service" DynamicDNSStatus="false" IsFailover="false">
            <Nameserver>dns1.registrar-servers.com</Nameserver>
            <Nameserver>dns2.registrar-servers.com</Nameserver>
          </DnsDetails>
        """
        extra_payload = {"DomainName": self.domain}

        try:
            xml = self.client.call("namecheap.domains.getInfo", extra_payload)
        except _ApiError as err:
            # this will happen if there is an API connection error
            # OR if the user is not permissioned to manage this domain
            # OR the API request came from a not whitelisted IP
            # we should print the error, so people know how to correct it.
            raise AuthenticationError(f"Authentication failed: `{str(err)}`")

        xpath = ".//{%(ns)s}CommandResponse/{%(ns)s}DomainGetInfoResult" % {
            "ns": _NAMESPACE
        }
        domain_info = xml.find(xpath)

        def _check_hosts_permission():
            # this shouldn't happen
            if domain_info is None:
                return False

            # do they own the domain?
            if domain_info.attrib["IsOwner"].lower() == "true":
                return True

            # look for rights
            xpath_alt = (
                ".//{%(ns)s}CommandResponse/{%(ns)s}DomainGetInfoResult"
                "/{%(ns)s}Modificationrights" % {"ns": _NAMESPACE}
            )
            rights_info = xml.find(xpath_alt)
            if rights_info is None:
                return False

            # all rights
            if rights_info.attrib["All"].lower() == "true":
                return True

            for right in rights_info:
                if right.attrib["Type"].lower() == "hosts":
                    # we're only looking at hosts, so we can exit now
                    return right.text.lower() == "ok"

            return None

        authorized = _check_hosts_permission()
        if not authorized:
            raise Exception(
                f"The domain {self.domain} is not controlled by this Namecheap account"
            )

        # FIXME What is this for?
        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def option_ttl(self):
        """
        Parse the `options` for a TTL.
        Used as a distinct function for documentation.
        via https://www.namecheap.com/support/api/methods/domains-dns/set-hosts.aspx
            Time to live for all record types.
            Possible values: any value between 60 to 60000
            Default Value: 1800
        if a TTL is submitted on the commandline:
            this will parse, but not validate it
        if no TTL is submitted:
            it will be ignored, allowing the server-side default
        Please note:
            Namecheap appears to have an internal cache on records; even with a
            short TTL of 60 seconds, it may take 120 seconds for their DNS to
            propagate.
        """
        return self._get_lexicon_option("ttl")

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, rtype, name, content):
        record = {
            # required
            "Type": rtype,
            "Name": self._relative_name(name),
        }

        # MX records needd special treatment
        if rtype == "MX":
            mxpref, address = content.split()
            record.update({"MxPref": int(mxpref), "Address": address})
        else:
            record.update({"Address": content})

        # inject the ttl if specified
        option_ttl = self.option_ttl()
        if option_ttl:
            record["TTL"] = option_ttl
        # LOGGER.debug('create_record: %s', 'id' in payload)
        # return 'id' in payload
        self.client.domains_dns_add_host(self.domain, record)
        return True

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is
    # received.
    def list_records(self, rtype=None, name=None, content=None):
        return self._list_records_internal(rtype=rtype, name=name, content=content)

    def _list_records_internal(
        self, rtype=None, name=None, content=None, identifier=None
    ):
        records = []
        raw_records = self.client.domains_dns_get_hosts(self.domain)
        for record in raw_records:
            records.append(self._convert_to_lexicon(record))

        if identifier:
            records = [record for record in records if record["id"] == identifier]
        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            if name.endswith("."):
                name = name[:-1]
            records = [record for record in records if name in record["name"]]
        if content:
            records = [
                record
                for record in records
                if record["content"].lower() == content.lower()
            ]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        # Delete record if it exists
        self.delete_record(identifier, rtype, name, content)
        return self.create_record(rtype, name, content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        records = self._list_records_internal(
            rtype=rtype, name=name, content=content, identifier=identifier
        )
        for record in records:
            self.client.domains_dns_del_host(
                self.domain, self._convert_to_namecheap(record)
            )
        return True

    def _convert_to_namecheap(self, record):
        """converts from lexicon format record to namecheap format record,
        suitable to sending through the api to namecheap"""

        processed_record = {}

        name = record["name"]
        if name.endswith("."):
            name = name[:-1]

        short_name = name[: name.find(self.domain) - 1]

        if record["type"] == "MX":
            # MX records needs to have separate treatment.
            mxpref, address = record["content"].split()
            processed_record.update({"MxPref": mxpref, "Address": address})
        else:
            processed_record.update({"Address": record["content"]})

        processed_record.update(
            {
                "Type": record["type"],
                "Name": short_name,
                "TTL": record["ttl"],
                "HostId": record["id"],
            }
        )

        return processed_record

    def _convert_to_lexicon(self, record):
        """converts from namecheap raw record format to lexicon format record"""

        name = record["Name"]

        if not name.endswith(self.domain):
            name += f".{self.domain}"

        content = (
            f"{record['MXPref']} {record['Address']}"
            if record["Type"] == "MX"
            else record["Address"]
        )

        processed_record = {
            "type": record["Type"],
            "name": name,
            "ttl": record["TTL"],
            "content": content,
            "id": record["HostId"],
        }

        return processed_record

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used by Namecheap provider
        pass


# Code below is a borrowed and simplified version of the Namecheap Python API
# implementation available here: https://github.com/Bemmu/PyNamecheap


class _ApiError(Exception):
    def __init__(self, number, text):
        Exception.__init__(self, "%s - %s" % (number, text))
        self.number = number
        self.text = text


class _Api:
    def __init__(
        self,
        api_user: str,
        api_key: str,
        username: str,
        client_ip: str,
        sandbox: bool = True,
    ):
        self.api_user = api_user
        self.api_key = api_key
        self.username = username
        self.client_ip = client_ip

        if sandbox:
            self.endpoint = "https://api.sandbox.namecheap.com/xml.response"
        else:
            self.endpoint = "https://api.namecheap.com/xml.response"

    def call(self, command: str, extra_payload: Optional[Dict] = None) -> Element:
        if not extra_payload:
            extra_payload = {}
        payload, extra_payload = self._payload(command, extra_payload)
        xml = self._fetch_xml(payload, extra_payload)
        return xml

    def _payload(self, command: str, extra_payload: Dict) -> Tuple[Dict, Dict]:
        payload = {
            "ApiUser": self.api_user,
            "ApiKey": self.api_key,
            "UserName": self.username,
            "ClientIP": self.client_ip,
            "Command": command,
        }
        # Namecheap recommends to use HTTP POST method when setting more than 10 hostnames
        # https://www.namecheap.com/support/api/methods/domains-dns/set-hosts.aspx
        if len(extra_payload) < 10:
            payload.update(extra_payload)
            extra_payload = {}
        return payload, extra_payload

    def _fetch_xml(
        self, payload: Dict, extra_payload: Optional[Dict] = None
    ) -> Element:
        if extra_payload:
            response = requests.post(self.endpoint, params=payload, data=extra_payload)
        else:
            response = requests.post(self.endpoint, params=payload)

        if not 200 <= response.status_code <= 299:
            raise _ApiError("1", "Did not receive 200 (Ok) response")

        xml = fromstring(response.text)

        if xml.attrib["Status"] == "ERROR":
            # Response namespace must be prepended to tag names.
            xpath = ".//{%(ns)s}Errors/{%(ns)s}Error" % {"ns": _NAMESPACE}
            error = xml.find(xpath)
            if error is not None:
                raise _ApiError(error.attrib["Number"], error.text)
            else:
                raise _ApiError(0, "Unknown exception")

        return xml

    def domains_dns_get_hosts(self, domain):
        extracted = tldextract.extract(domain)
        sld = extracted.domain
        tld = extracted.suffix
        extra_payload = {"SLD": sld, "TLD": tld}
        xml = self.call("namecheap.domains.dns.getHosts", extra_payload)
        xpath = ".//{%(ns)s}CommandResponse/{%(ns)s}DomainDNSGetHostsResult/*" % {
            "ns": _NAMESPACE
        }
        results = []
        for host in xml.findall(xpath):
            results.append(host.attrib)
        return results

    def domains_dns_add_host(self, domain, host_record):
        host_records_remote = self.domains_dns_get_hosts(domain)

        print("Remote: %i" % len(host_records_remote))

        host_records_remote.append(host_record)
        host_records_remote = [_elements_names_fix(x) for x in host_records_remote]

        print("To set: %i" % len(host_records_remote))

        extra_payload = _list_of_dictionaries_to_numbered_payload(host_records_remote)
        extracted = tldextract.extract(domain)
        sld = extracted.domain
        tld = extracted.suffix
        extra_payload.update({"SLD": sld, "TLD": tld})
        self.call("namecheap.domains.dns.setHosts", extra_payload)

    def domains_dns_del_host(self, domain, host_record) -> None:
        host_records_remote = self.domains_dns_get_hosts(domain)

        print("Remote: %i" % len(host_records_remote))

        host_records_new = []
        for r in host_records_remote:
            cond_type = r["Type"] == host_record["Type"]
            cond_name = r["Name"] == host_record["Name"]
            cond_addr = r["Address"] == host_record["Address"]

            if cond_type and cond_name and cond_addr:
                # skipping this record as it is the one we want to delete
                pass
            else:
                host_records_new.append(r)

        host_records_new = [_elements_names_fix(x) for x in host_records_new]

        print("To set: %i" % len(host_records_new))

        # Check that we delete not more than 1 record at a time
        if len(host_records_remote) != len(host_records_new) + 1:
            sys.stderr.write(
                "Something went wrong while removing host record, delta > 1: %i -> %i, aborting API call.\n"
                % (len(host_records_remote), len(host_records_new))
            )
            return

        extra_payload = _list_of_dictionaries_to_numbered_payload(host_records_new)
        extracted = tldextract.extract(domain)
        sld = extracted.domain
        tld = extracted.suffix
        extra_payload.update({"SLD": sld, "TLD": tld})
        self.call("namecheap.domains.dns.setHosts", extra_payload)


def _elements_names_fix(host_record):
    conversion_map = [("Name", "HostName"), ("Type", "RecordType")]

    for field in conversion_map:
        # if source field exists
        if field[0] in host_record:
            # convert it to target field and delete old one
            host_record[field[1]] = host_record[field[0]]
            del host_record[field[0]]

    return host_record


def _list_of_dictionaries_to_numbered_payload(data):
    return dict(
        sum(
            [[(k + str(i + 1), v) for k, v in d.items()] for i, d in enumerate(data)],
            [],
        )
    )



import datetime
import hashlib
import json
from argparse import ArgumentParser
from typing import Dict, List, Optional

import requests

from lexicon.exceptions import AuthenticationError, LexiconError
from lexicon.interfaces import Provider as BaseProvider


def _filter_rtype(rtype: str, rec) -> bool:
    if rec["type"] == rtype:
        return True
    return False


def _filter_name(name: str, rec) -> bool:
    if rec["name"] == name:
        return True
    return False


def _filter_content(content: str, rec) -> bool:
    if rec["content"] == content:
        return True
    return False


class Provider(BaseProvider):
    @staticmethod
    def get_nameservers() -> List[str]:
        return ["wedos.net", "wedos.eu", "wedos.cz", "wedos.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username",
            help="specify email address for authentication",
        )
        parser.add_argument(
            "--auth-pass",
            help="specify password for WAPI",
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.wedos.com/wapi/json"

    @staticmethod
    def _auth_hash(login, password):
        passhash = hashlib.sha1(password.encode("utf8")).hexdigest()
        phrase = login + passhash + datetime.datetime.now().strftime("%H")
        return hashlib.sha1(phrase.encode("utf8")).hexdigest()

    def authenticate(self):
        payload = self._post(data=self._create_payload("dns-domains-list", ""))
        domains = payload["response"]["data"]["domain"]
        for record in domains:
            if record["name"] == self.domain:
                self.domain_id = record["name"]
                break
            else:
                raise AuthenticationError("No domain found")

        self.domain_id = self.domain

    def cleanup(self) -> None:
        pass

    def _create_payload(self, command, payload_data):
        data = {
            "request": {
                "user": self._get_provider_option("auth_username"),
                "auth": self._auth_hash(
                    self._get_provider_option("auth_username"),
                    self._get_provider_option("auth_pass"),
                ),
                "command": command,
                "data": payload_data,
            }
        }
        return {"request": json.dumps(data)}

    def create_record(self, rtype: str, name: str, content: str) -> bool:
        records = self.list_records(rtype, name, content)
        if len(records) == 1:
            return True
        data = {
            "type": rtype,
            "name": self._relative_name(name),
            "rdata": content,
            "domain": self.domain_id,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = self._post(data=self._create_payload("dns-row-add", data))
        code = payload["response"]["code"]
        if code == 1000:
            validation = self._commit_changes()
            if validation:
                return True
            else:
                raise LexiconError("Cannot commit changes")
        else:
            raise LexiconError("Cannot create records")

    def list_records(
        self,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> List[Dict]:
        data = self._create_payload("dns-rows-list", {"domain": self.domain_id})
        payload = self._post(data=data)
        records = []
        dns_records = payload["response"]["data"]["row"]
        for record in dns_records:
            processed_record = {
                "type": record["rdtype"],
                "name": self._full_name(record["name"]),
                "ttl": record["ttl"],
                "content": record["rdata"],
                "id": record["ID"],
            }
            records.append(processed_record)
        if rtype is not None:
            records = list(rec for rec in records if _filter_rtype(rtype, rec))
        if name is not None:
            records = list(
                rec for rec in records if _filter_name(self._full_name(name), rec)
            )
        if content is not None:
            records = list(rec for rec in records if _filter_content(content, rec))

        return records

    def update_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if not identifier:
            records = self.list_records(rtype, name, content)
            identifiers = [record["id"] for record in records]
        else:
            identifiers = [identifier]

        if len(identifiers) == 0:
            return True
        payloads = []
        for record_id in identifiers:
            data = {
                "type": rtype,
                "rdata": content,
                "domain": self.domain_id,
                "row_id": record_id,
            }
            if name:
                data["name"] = self._full_name(name)
            if self._get_lexicon_option("ttl"):
                data["ttl"] = self._get_lexicon_option("ttl")
            payloads.append(
                self._post(data=self._create_payload("dns-row-update", data))
            )

        if all(payload["response"]["code"] == 1000 for payload in payloads):
            validation = self._commit_changes()
            if validation:
                return True
            else:
                raise LexiconError("Cannot commit changes")
        else:
            raise LexiconError("Cannot update records")

    def delete_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if not identifier:
            records = self.list_records(rtype, name, content)
            identifiers = [record["id"] for record in records]
        else:
            identifiers = [identifier]

        if len(identifiers) == 0:
            return True
        payloads = []
        for record_id in identifiers:
            data = {"domain": self.domain_id, "row_id": record_id}
            payloads.append(
                self._post(data=self._create_payload("dns-row-delete", data))
            )

        if all(payload["response"]["code"] == 1000 for payload in payloads):
            validation = self._commit_changes()
            if validation:
                return True
            else:
                raise LexiconError("Cannot commit changes")
        else:
            raise LexiconError("Cannot delete records")

    def _commit_changes(self) -> bool:
        data = {"name": self.domain_id}

        payload = self._post(data=self._create_payload("dns-domain-commit", data))
        code = payload["response"]["code"]
        if code == 1000:
            return True
        else:
            return False

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}

        response = requests.request(
            action,
            self.api_endpoint,
            data=data,
        )

        response.raise_for_status()
        return response.json()



"""Module provider for Softlayer"""

import logging
from argparse import ArgumentParser
from typing import List

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

try:
    import SoftLayer  # type: ignore
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Softlayer"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["softlayer.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-api-key", help="specify API private key for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None

        username = self._get_provider_option("auth_username")
        api_key = self._get_provider_option("auth_api_key")

        if not username or not api_key:
            raise Exception("No username and/or api key was specified")

        sl_client = SoftLayer.create_client_from_env(username=username, api_key=api_key)
        self.sl_dns = SoftLayer.managers.dns.DNSManager(sl_client)

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider,
    # so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason,
    # of if the domain does not exist.
    def authenticate(self):
        domain = self.domain

        payload = self.sl_dns.resolve_ids(domain)

        if not payload:
            raise AuthenticationError("No domain found")
        if len(payload) > 1:
            raise AuthenticationError("Too many domains found. This should not happen")

        LOGGER.debug("domain id: %s", payload[0])
        self.domain_id = payload[0]

    def cleanup(self) -> None:
        pass

    # Create record. If record already exists with the same content, do nothing

    def create_record(self, rtype, name, content):
        records = self.list_records(rtype, name, content)
        if records:
            # Nothing to do, record already exists
            LOGGER.debug("create_record: already exists")
            return True

        name = self._relative_name(name)
        ttl = self._get_lexicon_option("ttl")
        payload = self.sl_dns.create_record(self.domain_id, name, rtype, content, ttl)

        LOGGER.debug("create_record: %s", payload)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.

    def list_records(self, rtype=None, name=None, content=None):
        ttl = None
        if name:
            name = self._relative_name(name)

        payload = self.sl_dns.get_records(self.domain_id, ttl, content, name, rtype)

        records = []
        for record in payload:
            processed_record = {
                "type": record["type"].upper(),
                "name": self._full_name(record["host"]),
                "ttl": record["ttl"],
                "content": record["data"],
                "id": record["id"],
            }
            records.append(processed_record)

        LOGGER.debug("list_records: %s", records)
        return records

    # Update a record.
    # If an identifier is specified, use it, otherwise do a lookup using type and name.

    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            else:
                raise Exception("Record identifier could not be found.")

        record = {"id": identifier}
        if rtype:
            record["type"] = rtype
        if name:
            record["host"] = self._relative_name(name)
        if content:
            record["data"] = content
        if self._get_lexicon_option("ttl"):
            record["ttl"] = self._get_lexicon_option("ttl")

        self.sl_dns.edit_record(record)

        LOGGER.debug("update_record: %s", record)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self.sl_dns.delete_record(record_id)

        LOGGER.debug("delete_record: %s", True)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used in Softlayer
        pass



"""Module provider for TransIP"""

from __future__ import annotations

import binascii
import json
import logging
import uuid
from argparse import ArgumentParser
from base64 import b64decode, b64encode
from json import JSONDecodeError
from typing import Any, Dict, List, Optional

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from lexicon.config import ConfigResolver
from lexicon.exceptions import LexiconError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

API_BASE_URL = "https://api.transip.nl/v6"


class Provider(BaseProvider):
    """
    Provider class for TransIP

    provider_options can be overwritten by a Provider to setup custom defaults.
    They will be overwritten by any options set via the CLI or Env.
    order is:

    """

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-api-key",
            help="specify the private key to use for API authentication, in PEM format: can be either "
            "the path of the key file (eg. /tmp/key.pem) or the base64 encoded content of this "
            "file prefixed by 'base64::' (eg. base64::eyJhbGciOyJ...)",
        )
        parser.add_argument(
            "--auth-key-is-global",
            action="store_true",
            help="set this flag is the private key used is a global key with no IP whitelist restriction",
        )

    def __init__(self, config: ConfigResolver):
        super(Provider, self).__init__(config)
        self.provider_name = "transip"

        private_key_conf = self._get_provider_option("auth_api_key") or ""
        if private_key_conf.startswith("base64::"):
            private_key_bytes = b64decode(private_key_conf.replace("base64::", ""))
        else:
            with open(
                private_key_conf,
                "rb",
            ) as file:
                private_key_bytes = file.read()

        self.private_key = load_pem_private_key(private_key_bytes, password=None)
        self.token: str | None = None

    def authenticate(self):
        request_body = {
            "login": self._get_provider_option("auth_username"),
            "nonce": uuid.uuid4().hex,
            "global_key": self._get_provider_option("auth_key_is_global") or False,
        }

        request_body_bytes = json.dumps(request_body).encode()

        signature = self.private_key.sign(
            request_body_bytes,
            padding.PKCS1v15(),
            hashes.SHA512(),
        )

        headers = {"Signature": b64encode(signature).decode()}

        response = requests.request(
            "POST", f"{API_BASE_URL}/auth", json=request_body, headers=headers
        )
        response.raise_for_status()

        self.token = response.json()["token"]

        data = self._get(f"/domains/{self.domain}")

        self.domain_id = data["domain"]["authCode"]

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype: str, name: str, content: str) -> bool:
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        identifier = Provider._identifier(
            {"type": rtype, "name": self._full_name(name), "content": content}
        )

        if any(
            record
            for record in self.list_records(rtype=rtype, name=name, content=content)
            if record["id"] == identifier
        ):
            LOGGER.debug("create_record (ignored, duplicate): %s", identifier)
            return True

        data = {
            "dnsEntry": {
                "type": rtype,
                "name": self._relative_name(name),
                "content": content,
                "expire": self._get_lexicon_option("ttl"),
            },
        }

        self._post(f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("create_record: %s", identifier)

        return True

    def list_records(
        self,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        data = self._get(f"/domains/{self.domain}/dns")

        records = []
        for entry in data["dnsEntries"]:
            record = {
                "type": entry["type"],
                "name": self._full_name(entry["name"]),
                "ttl": entry["expire"],
                "content": entry["content"],
            }
            record["id"] = Provider._identifier(record)
            records.append(record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def update_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if not identifier and (not rtype or not name):
            raise Exception("Error, identifier or rtype+name parameters are required.")

        if identifier:
            records = self.list_records()
            records_to_update = [
                record for record in records if record["id"] == identifier
            ]
        else:
            records_to_update = self.list_records(rtype=rtype, name=name)

        if not records_to_update:
            raise Exception(
                f"Error, could not find a record for given identifier: {identifier}"
            )

        if len(records_to_update) > 1:
            LOGGER.warning(
                "Warning, multiple records found for given parameters, "
                "only first one will be updated: %s",
                records_to_update,
            )

        record = records_to_update[0]

        # TransIP API is not designed to update one record out of several records
        # matching the same type+name (eg. multi-valued TXT entries).
        # To circumvent the limitation, we remove first the record to update, then
        # recreate it with the updated content.

        data = {
            "dnsEntry": {
                "type": record["type"],
                "name": self._relative_name(record["name"]),
                "content": record["content"],
                "expire": record["ttl"],
            },
        }

        self._request("DELETE", f"/domains/{self.domain}/dns", data=data)

        data["dnsEntry"]["content"] = content

        self._post(f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("update_record: %s", record["id"])

        return True

    def delete_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if identifier:
            records = self.list_records()
            records = [record for record in records if record["id"] == identifier]

            if not records:
                raise LexiconError(
                    f"Could not find a record matching the identifier provider: {identifier}"
                )
        else:
            records = self.list_records(rtype, name, content)

        for record in records:
            data = {
                "dnsEntry": {
                    "type": record["type"],
                    "name": self._relative_name(record["name"]),
                    "content": record["content"],
                    "expire": record["ttl"],
                },
            }

            self._request("DELETE", f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("delete_records: %s %s %s %s", identifier, rtype, name, content)

        return True

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: Optional[Dict] = None,
        query_params: Optional[Dict] = None,
    ) -> Optional[Dict[str, Any]]:
        response = requests.request(
            action,
            f"{API_BASE_URL}{url}",
            params=query_params,
            json=data,
            headers={"Authorization": f"Bearer {self.token}"},
        )

        response.raise_for_status()

        try:
            return response.json()
        except JSONDecodeError:
            return None

    @staticmethod
    def _identifier(record):
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        digest.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
        digest.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

        return binascii.hexlify(digest.finalize()).decode("utf-8")[0:7]



"""\
Module provider for Duck DNS

The Duck DNS public API does not provide the whole set of features
supported by Lexicon: it only has one GET endpoint, "/update", authenticated by
the token which is given when you register with one of the supported
OAuth providers; the API only supports A, AAAA and TXT records, and
there is no other property that can be changed, like the TTL for these records;
the response message is plain text with "OK" or "KO" and only adds minimal
information when you supply the "verbose" parameter.

The DNS implementation supports only one record of each type
for each registered subdomain, and the value is propagated to every subdomain
after that, e.g. queries for testlexicon.duckdns.org and
example.given.testlexicon.duckdns.org return the same value.

Quirks of the DNS implementation:
* The A and AAAA records can be created separately, but when you delete one
the other is also deleted.
* When the A or AAAA records exist the TXT record is always present. Even if
you delete it will still be present with the value "". The implementation of
_list_records and _delete_record does not handle this special case.\
"""

import logging
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

try:
    import dns.name
    import dns.resolver
except ImportError:
    pass

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for Duck DNS"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["ca-central-1.compute.amazonaws.com"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-token", help="specify the account token for authentication"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_url = "https://www.duckdns.org"

    @staticmethod
    def _get_duckdns_domain(domain):
        domain_dns = dns.name.from_text(domain)

        if len(domain_dns.labels) > 2 and b"duckdns" not in domain_dns.labels:
            raise Exception("{} is not a valid Duck DNS domain.".format(domain))

        if b"duckdns" in domain_dns.labels:
            # use only the third level domain, the one actually registered in Duck DNS
            domain_index = domain_dns.labels.index(b"duckdns") - 1
        else:
            # use the last name if it is relative
            domain_index = -2

        return domain_dns.labels[domain_index].decode("utf-8")

    @staticmethod
    def _get_duckdns_rtype_param(rtype):
        if rtype not in ["A", "AAAA", "TXT"]:
            raise Exception("Duck DNS only supports A, AAAA and TXT records.")

        if rtype == "A":
            return "ip"
        elif rtype == "AAAA":
            return "ipv6"
        elif rtype == "TXT":
            return "txt"

    @staticmethod
    def _get_dns_record(name, rtype):
        if rtype not in ["A", "AAAA", "TXT"]:
            raise Exception("Duck DNS only supports A, AAAA and TXT records.")

        try:
            dns_rrset = Provider._get_dns_rrset(name, rtype)
        except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN):
            return {}

        ttl = dns_rrset.ttl

        if rtype == "A":
            content = dns_rrset[0].address
        elif rtype == "AAAA":
            content = dns_rrset[0].address
        else:
            content = dns_rrset[0].strings[0].decode("utf-8")

        record = {
            "id": rtype,
            "type": rtype,
            "name": name,
            "ttl": ttl,
            "content": content,
        }
        return record

    @staticmethod
    def _get_dns_rrset(name, rtype):
        resolver = dns.resolver.Resolver()
        resolver.lifetime = 60
        resolver.timeout = 70
        return resolver.resolve(name, rtype).rrset

    def authenticate(self):
        if self._get_provider_option("auth_token") is None:
            raise AuthenticationError("Must provide account token")

    def cleanup(self) -> None:
        pass

    # Create record. If the record already exists with the same content, do nothing"
    def create_record(self, rtype, name, content):
        if self._get_lexicon_option("ttl"):
            LOGGER.warning(
                "create_record: Duck DNS does not support modifying the TTL, ignoring {}".format(
                    self._get_lexicon_option("ttl")
                )
            )

        if rtype not in ["A", "AAAA", "TXT"]:
            raise Exception("Duck DNS only supports A, AAAA and TXT records.")

        if rtype == "A" and content == "auto":
            duckdns_content = ""
        else:
            duckdns_content = content

        params = {
            "domains": Provider._get_duckdns_domain(self.domain),
            Provider._get_duckdns_rtype_param(rtype): duckdns_content,
        }

        result = self._api_call(params)

        LOGGER.debug("create_record: %s", result)
        return "OK" in result

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, rtype=None, name=None, content=None):
        duckdns_domain = Provider._get_duckdns_domain(self.domain)
        full_duckdns_domain = duckdns_domain + ".duckdns.org"

        records = [
            Provider._get_dns_record(full_duckdns_domain, rtype)
            for rtype in ["A", "AAAA", "TXT"]
        ]
        processed_records = [
            {
                "id": record["id"],
                "type": record["type"],
                "name": self._full_name(name) if name is not None else record["name"],
                "ttl": record["ttl"],
                "content": record["content"],
            }
            for record in records
            if "id" in record
        ]
        filtered_records = [
            record
            for record in processed_records
            if (
                (rtype is None or record["type"] == rtype)
                and (name is None or record["name"] == self._full_name(name))
                and (content is None or record["content"] == content)
            )
        ]

        LOGGER.debug("list_records: %s", filtered_records)
        return filtered_records

    # Create or update a record.
    def update_record(self, identifier, rtype=None, name=None, content=None):
        if self._get_lexicon_option("ttl"):
            LOGGER.warning(
                "update_record: Duck DNS does not support modifying the TTL, ignoring {}".format(
                    self._get_lexicon_option("ttl")
                )
            )

        if not identifier:
            identifier = self._get_record_identifier(rtype=rtype, name=name)

        if rtype == "A" and content == "auto":
            duckdns_content = ""
        else:
            duckdns_content = content

        params = {
            "domains": Provider._get_duckdns_domain(self.domain),
            Provider._get_duckdns_rtype_param(identifier): duckdns_content,
        }

        result = self._api_call(params)

        LOGGER.debug("update_record: %s", result)
        return "OK" in result

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            identifier = self._get_record_identifier(
                rtype=rtype, name=name, content=content, delete=True
            )
        if identifier is None:
            LOGGER.debug("delete_record: no record found")
            return True

        params = {
            "domains": Provider._get_duckdns_domain(self.domain),
            Provider._get_duckdns_rtype_param(identifier): "",
            "clear": "true",
        }
        result = self._api_call(params)

        LOGGER.debug("delete_record: %s", result)
        return "OK" in result

    # Helpers
    def _api_call(self, params):
        response = self._request(query_params=dict(params.items()))

        content = response.content.decode("utf-8")

        if "KO" in content:
            raise Exception("KO: {}".format(params))

        return content

    def _full_name(self, record_name):
        record_name = record_name.rstrip(".")
        full_domain = Provider._get_duckdns_domain(self.domain) + ".duckdns.org"

        if not record_name.endswith(full_domain):
            record_name = f"{record_name}.{full_domain}"
        return record_name

    def _get_record_identifier(self, rtype=None, name=None, content=None, delete=False):
        records = self.list_records(rtype=rtype, name=name, content=content)
        if len(records) == 1:
            return records[0]["id"]

        if delete and len(records) == 0:
            return None
        else:
            raise Exception("Unambiguous record could not be found.")

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if query_params is None:
            query_params = {}
        query_params["token"] = self._get_provider_option("auth_token")
        query_params["verbose"] = "true"
        response = requests.request(
            "GET",
            self.api_url + "/update",
            params=query_params,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response



"""
Lexicon Plesk Provider

Author: Jens Reimann, 2018

API Docs: https://docs.plesk.com/en-US/onyx/api-rpc
"""

import logging
from argparse import ArgumentParser
from collections import defaultdict
from typing import Dict, List, Optional
from xml.etree import cElementTree
from xml.etree.ElementTree import Element

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

PLEX_URL_SUFFIX = "/enterprise/control/agent.php"


class Provider(BaseProvider):
    """Provider class for Plesk"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument(
            "--auth-password", help="specify password for authentication"
        )
        parser.add_argument(
            "--plesk-server", help="specify URL to the Plesk Web UI, including the port"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = self._get_provider_option("plesk_server")

        if self.api_endpoint.endswith("/"):
            self.api_endpoint = self.api_endpoint[:-1]

        if not self.api_endpoint.endswith(PLEX_URL_SUFFIX):
            self.api_endpoint += PLEX_URL_SUFFIX

        self.domain_id = None

        self.username = self._get_provider_option("auth_username")
        assert self.username is not None

        self.password = self._get_provider_option("auth_password")
        assert self.password is not None

    def __simple_request(self, rtype, operation, req):
        response = self.__plesk_request({rtype: {operation: req}})[rtype][operation]

        result = response["result"]

        if isinstance(result, list):
            for record in result:
                if record["status"] == "error":
                    raise Exception(
                        f"API returned at least one error: {record['errtext']}"
                    )
        elif response["result"]["status"] == "error":
            errcode = response["result"]["errcode"]
            errtext = response["result"]["errtext"]
            raise Exception(f"API returned error: {errcode} ({errtext})")

        return response

    def __plesk_request(self, request):
        headers = {
            "Content-type": "text/xml",
            "HTTP_PRETTY_PRINT": "TRUE",
            "HTTP_AUTH_LOGIN": self.username,
            "HTTP_AUTH_PASSWD": self.password,
        }

        xml = f"""\
<?xml version="1.0" encoding="utf-8"?>
{cElementTree.tostring(_dict_to_etree({"packet": request}), encoding="unicode")}\
"""

        LOGGER.debug("Request: %s", xml)

        response = requests.post(
            self.api_endpoint,
            headers=headers,
            data=xml,
            auth=(self.username, self.password),
        )

        data = response.text

        LOGGER.debug("Response: %s", data)
        result = _etree_to_dict(cElementTree.XML(data))
        return result["packet"]

    def __find_site(self):
        return self.__simple_request(
            "site", "get", {"filter": {"name": self.domain}, "dataset": {}}
        )["result"]["id"]

    def authenticate(self):
        self.domain_id = self.__find_site()

        if self.domain_id is None:
            raise AuthenticationError("Domain not found")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        return self.__create_entry(rtype, name, content, None)

    def list_records(self, rtype=None, name=None, content=None):
        entries = self.__find_dns_entries(rtype, name, content)
        LOGGER.debug("list_records: %s", entries)
        return entries

    def update_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is None:
            entries = self.__find_dns_entries(rtype, name, None)
            LOGGER.debug("Entries found: %s", entries)

            if not entries:
                raise Exception("No entry found for updating")

            identifier = entries[0]["id"]
            entry = self.__get_dns_entry(identifier)

            ids = []
            for an_entry in entries:
                ids.append(an_entry["id"])

            self.__delete_dns_records_by_id(ids)

        else:
            entry = self.__get_dns_entry(identifier)
            self.__delete_dns_records_by_id([identifier])

        assert entry is not None

        LOGGER.debug("Updating: %s", entry)

        if rtype:
            entry["type"] = rtype
        if name:
            entry["host"] = name
        if content:
            entry["value"] = content

        return self.__create_entry(
            entry["type"], entry["host"], entry["value"], entry["opt"]
        )

    def __create_entry(self, rtype, host, value, opt):
        entries = self.__find_dns_entries(rtype, self._fqdn_name(host), value)

        if entries:
            return True  # already exists

        self.__simple_request(
            "dns",
            "add_rec",
            {
                "site-id": self.domain_id,
                "type": rtype,
                "host": self._relative_name(host),
                "value": value,
                "opt": opt,
            },
        )

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            self.__delete_dns_records_by_id([identifier])
            return True
        entries = self.__find_dns_entries(rtype, self._fqdn_name(name), content)
        ids = []

        for entry in entries:
            ids.append(entry["id"])

        self.__delete_dns_records_by_id(ids)
        return bool(ids)

    def __get_dns_entry(self, identifier):
        return self.__simple_request("dns", "get_rec", {"filter": {"id": identifier}})[
            "result"
        ]["data"]

    def __find_dns_entries(self, rtype=None, host=None, value=None):
        LOGGER.debug("Searching for: %s, %s, %s", rtype, host, value)

        if value and rtype and rtype in ["CNAME"]:
            LOGGER.debug("CNAME transformation")
            value = value.rstrip(".") + "."

        if host:
            host = self._fqdn_name(host)

        result = self.__simple_request(
            "dns", "get_rec", {"filter": {"site-id": self.domain_id}}
        )

        entries = []

        for record in result["result"]:
            LOGGER.debug("Record: %s", record)

            if (rtype is not None) and (record["data"]["type"] != rtype):
                LOGGER.debug(
                    "\tType doesn't match - expected: '%s', found: '%s'",
                    rtype,
                    record["data"]["type"],
                )
                continue

            if (host is not None) and (record["data"]["host"] != host):
                LOGGER.debug(
                    "\tHost doesn't match - expected: '%s', found: '%s'",
                    host,
                    record["data"]["host"],
                )
                continue

            if (value is not None) and (record["data"]["value"] != value):
                LOGGER.debug(
                    "\tValue doesn't match - expected: '%s', found: '%s'",
                    value,
                    record["data"]["value"],
                )
                continue

            entry = {
                "id": record["id"],
                "type": record["data"]["type"],
                "name": self._full_name(record["data"]["host"]),
                "ttl": None,
                "options": {},
            }

            if record["data"]["type"] in ["CNAME"]:
                entry["content"] = record["data"]["value"].rstrip(".")
            else:
                entry["content"] = record["data"]["value"]

            if record["data"]["type"] == "MX":
                entry["options"]["mx"] = {"priority": int(record["data"]["opt"])}

            entries.append(entry)

        return entries

    def __delete_dns_records_by_id(self, ids):
        if not ids:
            return

        req = []
        for i in ids:
            req.append({"del_rec": {"filter": {"id": i}}})

        self.__plesk_request({"dns": req})

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Helper _request is not used for Plesk provider
        pass


def _etree_to_dict(t: Element) -> Optional[Dict]:
    d: Optional[Dict] = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(_etree_to_dict, children):
            if dc:
                for k, v in dc.items():
                    dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib and d:
        d[t.tag].update(("@" + k, v) for k, v in t.attrib.items())
    if t.text and d:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]["#text"] = text
        else:
            d[t.tag] = text
    return d


def _dict_to_etree(d: Dict) -> Element:
    def _to_etree(d1, root):
        if not d1:
            pass
        elif isinstance(d1, str):
            root.text = d1
        elif isinstance(d1, dict):
            for k, v in d1.items():
                assert isinstance(k, str)
                if k.startswith("#"):
                    assert k == "#text" and isinstance(v, str)
                    root.text = v
                elif k.startswith("@"):
                    assert isinstance(v, str)
                    root.set(k[1:], v)
                elif isinstance(v, list):
                    for e in v:
                        _to_etree(e, cElementTree.SubElement(root, k))
                else:
                    _to_etree(v, cElementTree.SubElement(root, k))
        else:
            raise TypeError("invalid type: " + str(type(d1)))

    assert isinstance(d, dict) and len(d) == 1
    tag, body = next(iter(d.items()))
    node = cElementTree.Element(tag)
    _to_etree(body, node)
    return node



"""Module provider for DirectAdmin hosts"""

import logging
import warnings
from argparse import ArgumentParser
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


# See https://www.directadmin.com/features.php?id=504 for the specification of
# the URIs for the different operations
class Provider(BaseProvider):
    """Provider class for DirectAdmin"""

    @staticmethod
    def get_nameservers() -> List[str]:
        # DirectAdmin is not tied to a specific domain, so there is nothing to specify here
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            "--auth-password",
            help="specify password for authentication (or login key for two-factor authentication)",
        )
        parser.add_argument(
            "--auth-username", help="specify username for authentication"
        )
        parser.add_argument("--endpoint", help="specify the DirectAdmin endpoint")

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.endpoint = self._get_provider_option("endpoint")
        if self.endpoint is None:
            raise Exception("Specify endpoint of DirectAdmin")

    def authenticate(self):
        try:
            response = self._get("/CMD_API_SHOW_DOMAINS")
        except requests.exceptions.HTTPError as err:
            # A 401 error will be returned in case of incorrect or missing
            # credentials
            raise err

        # The response is a URL encoded array containing all available domains
        domains = [domain.split("=").pop() for domain in response.split("&")]
        try:
            self.domain_id = domains.index(self.domain)
        except ValueError:
            raise AuthenticationError(f"Domain {self.domain} not found")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        # Refuse to create duplicate records
        existing_records = self.list_records(rtype, name, content)
        if existing_records:
            return True

        query_params = {
            "action": "add",
            "json": "yes",
            "name": f"{self._full_name(name)}.",
            "type": rtype,
            "value": content,
        }

        if self._get_lexicon_option("ttl"):
            query_params["ttl"] = self._get_lexicon_option("ttl")

        try:
            response = self._get("/CMD_API_DNS_CONTROL", query_params)
        except requests.exceptions.HTTPError:
            response = {"success": "Create Failed"}

        LOGGER.debug("create_record: %s", response)

        return response["success"].lower().find("added") >= 0

    def list_records(self, rtype=None, name=None, content=None):
        response = {"records": []}
        try:
            response = self._get("/CMD_API_DNS_CONTROL", {"json": "yes"})
        except requests.exceptions.HTTPError as err:
            warnings.warn(err.response.text)
            raise

        records = [
            self._parse_response_record(record) for record in response["records"]
        ]
        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            cmp_name = self._full_name(name.lower())
            records = [record for record in records if record["name"] == cmp_name]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def _parse_response_record(self, response_record):
        # Most fields in a record response match, except for content which is
        # value
        return {
            "content": response_record["value"],
            "id": response_record["combined"],
            "name": self._full_name(response_record["name"]),
            # 4 hours appears to be the default shown by DirectAdmin if no TTL
            # is set
            "ttl": response_record.get("ttl", 14400),
            "type": response_record["type"],
        }

    def update_record(self, identifier, rtype=None, name=None, content=None):
        # Editing a record is a combination of removing the old record and
        # adding a new one, but specifying 'edit' as the action while passing
        # all parameters necessary for both deletion and creation

        if not identifier:
            # The identifier of the original record to edit is necessary to be
            # able to create the appropriate delete payload
            original_records = self.list_records(rtype, name)
            if len(original_records) > 1:
                warnings.warn("Found multiple records to edited. Cannot continue...")
                return False

            if not original_records:
                warnings.warn("Found no records to edit. Cannot continue...")
                return False

            identifier = original_records[0]["id"]

        delete_key = self._determine_delete_key(identifier, rtype)

        query_params = {
            "action": "edit",
            "json": "yes",
            delete_key: identifier,
            "name": f"{self._full_name(name)}.",
            "type": rtype,
            "value": content,
        }

        try:
            response = self._get("/CMD_API_DNS_CONTROL", query_params)
        except requests.exceptions.HTTPError:
            response = {"success": "Update Failed"}

        LOGGER.debug("update_record: %s", response)

        return response["success"].lower().find("added") > 0

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            identifiers = [
                record["id"] for record in self.list_records(rtype, name, content)
            ]
        else:
            identifiers = [identifier]

        response = {"success": "noop"}
        for identifier_to_delete in identifiers:
            delete_key = self._determine_delete_key(identifier_to_delete, rtype)
            query_params = {
                "action": "select",
                "json": "yes",
                delete_key: identifier_to_delete,
            }

            try:
                response = self._get("/CMD_API_DNS_CONTROL", query_params)
            except requests.exceptions.HTTPError:
                response = {"success": "Delete Failed"}

            LOGGER.debug("delete_record: %s", response)

        return response["success"].lower().find("deleted") > 0

    def _determine_delete_key(self, identifier, rtype):
        # The indicator for the record that needs to be removed is determined
        # by the type of the record and its index within all records of that
        # type. There's an additional check on the name and value which still
        # need to match for the removal to actually occur
        if not rtype:
            # An rtype is necessary to create the delete_key. However, it may
            # not be specified in which case an effort needs to be made to
            # figure it out automatically. The necessary data can be recreated
            # from the identifier
            identifier_parts = identifier.split("&")
            name = identifier_parts[0].split("=").pop()
            content = identifier_parts[1].split("=").pop()

            records = self.list_records(None, name, content)
            rtype = records[0]["type"]

        existing_records = self.list_records(rtype)
        existing_record_index = 0
        for index, record in enumerate(existing_records):
            if record["id"] == identifier:
                existing_record_index = index

        return f"{rtype}recs{existing_record_index}".lower()

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params["domain"] = self.domain

        response = requests.request(
            action,
            self.endpoint + url,
            auth=HTTPBasicAuth(
                self._get_provider_option("auth_username"),
                self._get_provider_option("auth_password"),
            ),
            params=query_params,
        )

        # If the request fails for any reason, throw an error
        response.raise_for_status()

        if "json" in response.headers["Content-Type"].lower():
            return response.json()

        return response.text



"""Module provider for OnApp"""

import json
import logging
from argparse import ArgumentParser
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)


class Provider(BaseProvider):
    """Provider class for OnApp"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return []

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            The OnApp provider requires your OnApp account\'s email address and
            API token, which can be found on your /profile page on the Control Panel interface.
            The server is your dashboard URL, with format like https://dashboard.youronapphost.org"""
        parser.add_argument(
            "--auth-username", help="specify email address of the OnApp account"
        )
        parser.add_argument(
            "--auth-token", help="specify API Key for the OnApp account"
        )
        parser.add_argument(
            "--auth-server", help="specify URL to the OnApp Control Panel Server"
        )

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.domain_id = None

        if not self._get_provider_option("auth_username"):
            raise Exception("Error, OnApp Email Address is not defined")
        if not self._get_provider_option("auth_token"):
            raise Exception("Error, OnApp API Key is not defined")
        if not self._get_provider_option("auth_server"):
            raise Exception("Error, OnApp Control Panel URL is not defined")

        self.session = requests.Session()

    def authenticate(self):
        domain = self.domain

        zones = self._get("/dns_zones.json")
        for zone in zones:
            if zone["dns_zone"]["name"] == domain:
                self.domain_id = zone["dns_zone"]["id"]
                break
        else:
            raise AuthenticationError(f"Could not find {domain} in OnApp DNS Zones")

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        data = {
            "name": self._relative_name(name),
            "type": rtype,
            self._key_for_record_type(rtype): content,
        }

        ttl = self._get_lexicon_option("ttl")
        if ttl:
            data["ttl"] = f"{ttl}"

        result = self._post(
            f"/dns_zones/{self.domain_id}/records.json", {"dns_record": data}
        )
        LOGGER.debug("create_record: %s", result)

        return True

    def list_records(self, rtype=None, name=None, content=None):
        records = []

        response = self._get(f"/dns_zones/{self.domain_id}/records.json")
        for record_type in response["dns_zone"]["records"]:
            # For now we do not support other RR types so we ignore them, also see
            # _key_for_record_type
            if record_type not in ("A", "AAAA", "CNAME", "TXT"):
                continue

            if rtype and record_type != rtype:
                continue

            for record in response["dns_zone"]["records"][record_type]:
                record = record["dns_record"]

                if name and record["name"] != self._relative_name(name):
                    continue

                record_content = record[self._key_for_record_type(record_type)]
                if content and record_content != content:
                    continue

                records.append(
                    {
                        "id": record["id"],
                        "name": self._full_name(record["name"]),
                        "type": record["type"],
                        "ttl": record["ttl"],
                        "content": record_content,
                    }
                )

        LOGGER.debug("list_records: %s", records)

        return records

    def update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            existing = self._guess_record(rtype, name)
            identifier = existing["id"]

        ttl = self._get_lexicon_option("ttl")

        if not name or not ttl:
            if not existing:
                existing = self._get(
                    f"/dns_zones/{self.domain_id}/records/{identifier}.json"
                )
            if not name:
                name = existing["name"]
            if not ttl:
                ttl = existing["ttl"]

        request = {
            "name": self._relative_name(name),
            "ttl": f"{ttl}",
            self._key_for_record_type(rtype): content,
        }

        result = self._put(
            f"/dns_zones/{self.domain_id}/records/{identifier}.json",
            {"dns_record": request},
        )
        LOGGER.debug("update_record: %s", result)

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        deletion_ids = []

        if not identifier:
            records = self.list_records(rtype, name, content)
            deletion_ids = [record["id"] for record in records]
        else:
            deletion_ids.append(identifier)

        for one_id in deletion_ids:
            self._delete(f"/dns_zones/{self.domain_id}/records/{one_id}.json")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        target = self._get_provider_option("auth_server") + url

        body = ""
        if data is not None:
            body = json.dumps(data)

        auth = HTTPBasicAuth(
            self._get_provider_option("auth_username"),
            self._get_provider_option("auth_token"),
        )

        request = requests.Request(
            action, target, data=body, headers=headers, params=query_params, auth=auth
        )
        prepared_request = self.session.prepare_request(request)

        result = self.session.send(prepared_request)
        result.raise_for_status()

        if result.text:
            return result.json()
        return None

    def _key_for_record_type(self, record_type):
        if record_type in ("A", "AAAA"):
            return "ip"
        if record_type == "CNAME":
            return "hostname"
        if record_type == "TXT":
            return "txt"
        if record_type in ("MX", "NS", "SOA", "SRV", "LOC"):
            raise Exception(
                f"{record_type} record type is not supported in the OnApp Provider"
            )
        raise Exception(f"{record_type} record type is unknown")

    def _guess_record(self, rtype, name=None, content=None):
        records = self.list_records(rtype=rtype, name=name, content=content)
        if len(records) == 1:
            return records[0]
        if len(records) > 1:
            raise Exception(
                "Identifier was not provided and several existing records "
                "match the request for {0}/{1}".format(rtype, name)
            )
        raise Exception(
            "Identifier was not provided and no existing records "
            "match the request for {0}/{1}".format(rtype, name)
        )



"""Module provider for OVH"""

import hashlib
import json
import logging
import time
from argparse import ArgumentParser
from typing import List

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.interfaces import Provider as BaseProvider

try:
    from simplejson import JSONDecodeError  # type: ignore
except ImportError:
    from json import JSONDecodeError

LOGGER = logging.getLogger(__name__)

ENDPOINTS = {
    "ovh-eu": "https://eu.api.ovh.com/1.0",
    "ovh-ca": "https://ca.api.ovh.com/1.0",
    "ovh-us": "https://api.ovhcloud.com/1.0",
    "kimsufi-eu": "https://eu.api.kimsufi.com/1.0",
    "kimsufi-ca": "https://ca.api.kimsufi.com/1.0",
    "soyoustart-eu": "https://eu.api.soyoustart.com/1.0",
    "soyoustart-ca": "https://ca.api.soyoustart.com/1.0",
}


class Provider(BaseProvider):
    """Provider class for OVH"""

    @staticmethod
    def get_nameservers() -> List[str]:
        return ["ovh.net", "anycast.me"]

    @staticmethod
    def configure_parser(parser: ArgumentParser) -> None:
        parser.description = """
            OVH Provider requires a token with full rights on /domain/*.
            It can be generated for your OVH account on the following URL:
            https://api.ovh.com/createToken/index.cgi?GET=/domain/*&PUT=/domain/*&POST=/domain/*&DELETE=/domain/*"""
        parser.add_argument(
            "--auth-entrypoint",
            help="specify the OVH entrypoint",
            choices=[
                "ovh-eu",
                "ovh-ca",
                "soyoustart-eu",
                "soyoustart-ca",
                "kimsufi-eu",
                "kimsufi-ca",
            ],
        )
        parser.add_argument(
            "--auth-application-key", help="specify the application key"
        )
        parser.add_argument(
            "--auth-application-secret", help="specify the application secret"
        )
        parser.add_argument("--auth-consumer-key", help="specify the consumer key")

    def __init__(self, config):
        super(Provider, self).__init__(config)

        # Handling missing required parameters
        if not self._get_provider_option("auth_entrypoint"):
            raise Exception("Error, entrypoint is not defined")
        if not self._get_provider_option("auth_application_key"):
            raise Exception("Error, application key is not defined")
        if not self._get_provider_option("auth_application_secret"):
            raise Exception("Error, application secret is not defined")
        if not self._get_provider_option("auth_consumer_key"):
            raise Exception("Error, consumer key is not defined")

        # Construct DNS OVH environment
        self.domain_id = None
        self.endpoint_api = ENDPOINTS.get(self._get_provider_option("auth_entrypoint"))
        self.session = None
        self.time_delta = None

    def authenticate(self):
        # All requests will be done in one HTTPS session
        self.session = requests.Session()

        # Calculate delta time between local and OVH to avoid requests rejection
        server_time = self.session.get(f"{self.endpoint_api}/auth/time").json()
        self.time_delta = server_time - int(time.time())

        # Get domain and status
        domain = self.domain

        domains = self._get("/domain/zone/")
        if domain not in domains:
            raise AuthenticationError(f"Domain {domain} not found")

        status = self._get(f"/domain/zone/{domain}/status")
        if not status["isDeployed"]:
            raise AuthenticationError(f"Zone {domain} is not deployed")

        self.domain_id = domain

    def cleanup(self) -> None:
        pass

    def create_record(self, rtype, name, content):
        domain = self.domain
        ttl = self._get_lexicon_option("ttl")

        records = self.list_records(rtype, name, content)
        for record in records:
            if (
                record["type"] == rtype
                and self._relative_name(record["name"]) == self._relative_name(name)
                and record["content"] == content
            ):
                LOGGER.debug(
                    "create_record (ignored, duplicate): %s %s %s", rtype, name, content
                )
                return True

        data = {
            "fieldType": rtype,
            "subDomain": self._relative_name(name),
            "target": content,
        }

        if ttl:
            data["ttl"] = ttl

        result = self._post(f"/domain/zone/{domain}/record", data)
        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("create_record: %s", result["id"])

        return True

    def list_records(self, rtype=None, name=None, content=None):
        domain = self.domain
        records = []

        params = {}
        if rtype:
            params["fieldType"] = rtype
        if name:
            params["subDomain"] = self._relative_name(name)

        record_ids = self._get(f"/domain/zone/{domain}/record", params)

        for record_id in record_ids:
            raw = self._get(f"/domain/zone/{domain}/record/{record_id}")
            records.append(
                {
                    "type": raw["fieldType"],
                    "name": self._full_name(raw["subDomain"]),
                    "ttl": raw["ttl"],
                    "content": raw["target"],
                    "id": raw["id"],
                }
            )

        if content:
            records = [
                record
                for record in records
                if record["content"].lower()
                in {content.lower(), f'"{content.lower()}"'}
            ]

        LOGGER.debug("list_records: %s", records)

        return records

    def update_record(self, identifier, rtype=None, name=None, content=None):
        domain = self.domain

        if not identifier:
            records = self.list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) > 1:
                raise Exception("Several record identifiers match the request")
            else:
                raise Exception("Record identifier could not be found")

        data = {}
        if name:
            data["subDomain"] = self._relative_name(name)
        if content:
            data["target"] = content

        self._put(f"/domain/zone/{domain}/record/{identifier}", data)
        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("update_record: %s", identifier)

        return True

    def delete_record(self, identifier=None, rtype=None, name=None, content=None):
        domain = self.domain

        delete_record_id = []
        if not identifier:
            records = self.list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/domain/zone/{domain}/record/{record_id}")

        self._post(f"/domain/zone/{domain}/refresh")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {}
        target = self.endpoint_api + url
        body = ""

        if data is not None:
            headers["Content-type"] = "application/json"
            body = json.dumps(data)

        # Get correctly sync time
        now = str(int(time.time()) + self.time_delta)

        headers["X-Ovh-Application"] = self._get_provider_option("auth_application_key")
        headers["X-Ovh-Consumer"] = self._get_provider_option("auth_consumer_key")
        headers["X-Ovh-Timestamp"] = now

        request = requests.Request(
            action, target, data=body, params=query_params, headers=headers
        )
        prepared_request = self.session.prepare_request(request)

        # Build OVH API signature for the current request
        signature = hashlib.sha1()
        signature.update(
            "+".join(
                [
                    self._get_provider_option("auth_application_secret"),
                    self._get_provider_option("auth_consumer_key"),
                    action.upper(),
                    prepared_request.url,
                    body,
                    now,
                ]
            ).encode("utf-8")
        )

        # Sign the request
        headers["X-Ovh-Signature"] = "$1$" + signature.hexdigest()

        result = self.session.request(
            method=action, url=target, params=query_params, data=body, headers=headers
        )
        result.raise_for_status()

        try:
            return result.json()
        except JSONDecodeError:
            LOGGER.warning(
                f"Unexpected response from OVH APIs for {action} {url} (response dumped as plain text):\n{result.text}"
            )
            return None
